#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This is the direction-screen operation for LoSoTo

from losoto.lib_operations import *
from losoto.operations.stationscreen import _getxy, _radec2xy, _xy2radec, _makeWCS
from losoto.operations.stationscreen import _flag_outliers, _circ_chi2
from losoto._logging import logger as logging

logging.debug('Loading DIRECTIONSCREEN module.')


def _run_parser(soltab, parser, step):
    outSoltab = parser.getstr( step, "outSoltab", 'tecscreen' )
    height = parser.getfloat( step, "height", 200e3 )
    order = parser.getint( step, "Order", 5 )
    ncpu = parser.getint( '_global', "npcu", 0 )

    parser.checkSpelling( step, soltab, ['outSoltab', 'height', 'order'])
    return run(soltab, outSoltab, height, order, ncpu)


def _calculate_piercepoints(station_positions, source_positions, times, height=200e3):
    """
    Returns array of piercepoint locations and airmass values for a
    screen at the given height (in m)

    Parameters
    ----------
    station_positions : array
        array of station positions
    source_positions : array
        array of source positions
    times: array
        array of times
    height: float
        height of screen (m)

    Returns
    -------
    pp: array
        array of pierce points
    airmass: array
        array of airmass values
    midRA: float
        reference RA for WCS system (deg)
    midDec: float
        reference Dec for WCS system (deg)

    """
    import pyrap.measures
    import numpy as np
    try:
        import progressbar
    except ImportError:
        import losoto.progressbar as progressbar

    logging.info('Calculating screen pierce-point locations and airmass values...')
    N_sources = source_positions.shape[0]
    N_stations = station_positions.shape[0]
    N_piercepoints = N_stations * N_sources
    N_times = len(times)
    pp = np.zeros((N_times, N_piercepoints, 3))
    airmass = np.zeros((N_times, N_piercepoints))

    me = pyrap.measures.measures()
    position = me.position('ITRF', '%fm' % station_positions[0,0],
        '%fm' % station_positions[0,1], '%fm' % station_positions[0,2])
    me.doframe(position)

    pbar = progressbar.ProgressBar(maxval=N_times).start()
    ipbar = 0
    for k in range(N_times):
        epoch = me.epoch('UTC', '%fs' % times[k])
        me.doframe(epoch)
        pp_idx = 0
        for i in range(N_sources):
            ra = source_positions[i,0]
            dec = source_positions[i,1]
            d = me.direction('J2000', '%frad' % ra, '%frad' % dec)
            d1 = me.measure(d, 'ITRF')
            phi = d1['m0']['value']
            theta = d1['m1']['value']
            dx = np.cos(theta)*np.cos(phi)
            dy = np.cos(theta)*np.sin(phi)
            dz = np.sin(theta)
            direction = np.array((dx,dy,dz))
            for station_position in station_positions:
                pp[k, pp_idx, :], airmass[k, pp_idx] = _calc_piercepoint(
                    station_position, direction, height)
                pp_idx += 1
        pbar.update(ipbar)
        ipbar += 1
    pbar.finish()
    midRA = 0.0
    midDec = 0.0

    return pp, airmass, midRA, midDec


def _calc_piercepoint(pos, direction, height):
    """
    Calculates pierce point locations and airmass values for given station
    position, source direction, and height (in m)

    Parameters
    ----------
    pos: array
        array of station positions
    direction: array
        array of source directions
    height: float
        height of screen (m)

    Returns
    -------
    pp: array
        array of pierce points
    airmass: array
        array of airmass values

    """
    import numpy as np

    pp = np.zeros(3)
    earth_ellipsoid_a = 6378137.0
    earth_ellipsoid_a2 = earth_ellipsoid_a * earth_ellipsoid_a
    earth_ellipsoid_b = 6356752.3142
    earth_ellipsoid_b2 = earth_ellipsoid_b * earth_ellipsoid_b
    earth_ellipsoid_e2 = (earth_ellipsoid_a2 - earth_ellipsoid_b2) / earth_ellipsoid_a2;

    ion_ellipsoid_a = earth_ellipsoid_a + height
    ion_ellipsoid_a2_inv = 1.0 / (ion_ellipsoid_a * ion_ellipsoid_a)
    ion_ellipsoid_b = earth_ellipsoid_b + height
    ion_ellipsoid_b2_inv = 1.0 / (ion_ellipsoid_b * ion_ellipsoid_b)

    x = pos[0] / ion_ellipsoid_a
    y = pos[1] / ion_ellipsoid_a
    z = pos[2] / ion_ellipsoid_b
    c = x*x + y*y + z*z - 1.0

    dx = direction[0] / ion_ellipsoid_a
    dy = direction[1] / ion_ellipsoid_a
    dz = direction[2] / ion_ellipsoid_b
    a = dx*dx + dy*dy + dz*dz
    b = x*dx + y*dy  + z*dz
    alpha = (-b + np.sqrt(b*b - a*c)) / a
    pp = pos[0] + alpha * direction
    normal_x = pp[0] * ion_ellipsoid_a2_inv
    normal_y = pp[1] * ion_ellipsoid_a2_inv
    normal_z = pp[2] * ion_ellipsoid_b2_inv
    norm_normal2 = normal_x*normal_x + normal_y*normal_y + normal_z*normal_z
    norm_normal = np.sqrt(norm_normal2)
    sin_lat2 = normal_z*normal_z / norm_normal2
    g = 1.0 - earth_ellipsoid_e2 * sin_lat2
    sqrt_g = np.sqrt(g)
    M = earth_ellipsoid_b2 / ( earth_ellipsoid_a * g * sqrt_g )
    N = earth_ellipsoid_a / sqrt_g

    local_ion_ellipsoid_e2 = (M - N) / ((M + height) * sin_lat2 - N - height);
    local_ion_ellipsoid_a = (N + height) * np.sqrt(1.0 - local_ion_ellipsoid_e2 * sin_lat2)
    local_ion_ellipsoid_b = local_ion_ellipsoid_a * np.sqrt(1.0 - local_ion_ellipsoid_e2)

    z_offset = ((1.0 - earth_ellipsoid_e2) * N + height - (1.0 -
        local_ion_ellipsoid_e2) * (N+height)) * np.sqrt(sin_lat2)

    x1 = pos[0] / local_ion_ellipsoid_a
    y1 = pos[1] / local_ion_ellipsoid_a
    z1 = (pos[2] - z_offset) / local_ion_ellipsoid_b
    c1 = x1*x1 + y1*y1 + z1*z1 - 1.0

    dx = direction[0] / local_ion_ellipsoid_a
    dy = direction[1] / local_ion_ellipsoid_a
    dz = direction[2] / local_ion_ellipsoid_b
    a = dx*dx + dy*dy + dz*dz
    b = x1*dx + y1*dy + z1*dz
    alpha = (-b + np.sqrt(b*b - a*c1)) / a

    pp = pos + alpha * direction

    normal_x = pp[0] * ion_ellipsoid_a2_inv
    normal_y = pp[1] * ion_ellipsoid_a2_inv
    normal_z = (pp[2] - z_offset) * ion_ellipsoid_b2_inv

    norm_normal2 = normal_x*normal_x + normal_y*normal_y + normal_z*normal_z
    norm_normal = np.sqrt(norm_normal2)

    airmass = norm_normal / (direction[0]*normal_x + direction[1]*normal_y +
        direction[2]*normal_z)

    return pp, airmass


def _fit_phase_screen(station_names, source_names, pp, airmass, rr, weights, times,
    height, order, r_0, beta, outQueue):
    """
    Fits a screen to given phase values using Karhunen-Lo`eve base vectors

    Parameters
    ----------
    station_names: array
        Array of station names
    source_names: array
        Array of source names
    pp: array
        Array of piercepoint locations
    airmass: array
        Array of airmass values (note: not currently used)
    rr: array
        Array of phase values to fit screen to
    weights: array
        Array of weights
    times: array
        Array of times
    height: float
        Height of screen (m)
    order: int
        Order of screen (i.e., number of KL base vectors to keep)
    r_0: float
        Scale size of phase fluctuations (m)
    beta: float
        Power-law index for phase structure function (5/3 => pure Kolmogorov
        turbulence)

    """
    import numpy as np
    from pylab import kron, concatenate, pinv, norm, newaxis, find, amin, svd, eye

    logging.info('Fitting screens...')

    # Initialize arrays
    N_stations = len(station_names)
    N_sources = len(source_names)
    N_times = len(times)
    N_piercepoints = N_sources * N_stations
    real_fit_white_all = np.zeros((N_times, N_sources, N_stations))
    imag_fit_white_all = np.zeros((N_times, N_sources, N_stations))
    phase_fit_white_all = np.zeros((N_times, N_sources, N_stations))
    real_residual_all = np.zeros((N_times, N_sources, N_stations))
    imag_residual_all = np.zeros((N_times, N_sources, N_stations))
    phase_residual_all = np.zeros((N_times, N_sources, N_stations))

    # Change phase to real/imag
    rr_real = np.cos(rr)
    rr_imag = np.sin(rr)

    for k in range(N_times):
        try:
            D = np.resize(pp[k, :, :], (N_piercepoints, N_piercepoints, 3))
            D = np.transpose(D, (1, 0, 2)) - D
            D2 = np.sum(D**2, axis=2)
            C = -(D2 / r_0**2)**(beta / 2.0) / 2.0
            pinvC = pinv(C, rcond=1e-3)
            U, S, V = svd(C)
            invU = pinv(np.dot(np.transpose(U[:, :order]), np.dot(weights[:, :, k], U[:, :order])), rcond=1e-3)

            # Calculate real screen
            rr1 = np.dot(np.transpose(U[:, :order]), np.dot(weights[:, :, k], rr_real[:, k]))
            real_fit = np.dot(pinvC, np.dot(U[:, :order], np.dot(invU, rr1)))
            real_fit_white_all[k, :, :] = real_fit.reshape((N_sources, N_stations))
            residual = rr_real - np.dot(C, real_fit)[:, newaxis]
            real_residual_all[k, :, :] = residual.reshape((N_sources, N_stations))

            # Calculate imag screen
            rr1 = np.dot(np.transpose(U[:, :order]), np.dot(weights[:, :, k], rr_imag[:, k]))
            imag_fit = np.dot(pinvC, np.dot(U[:, :order], np.dot(invU, rr1)))
            imag_fit_white_all[k, :, :] = imag_fit.reshape((N_sources, N_stations))
            residual = rr_imag - np.dot(C, imag_fit)[:, newaxis]
            imag_residual_all[k, :, :] = residual.reshape((N_sources, N_stations))

            # Calculate phase screen
            phase_fit = np.dot(pinvC, np.arctan2(np.dot(C, imag_fit), np.dot(C, real_fit)))
            phase_fit_white_all[k, :, :] = phase_fit.reshape((N_sources, N_stations))
            residual = rr - np.dot(C, phase_fit)[:, newaxis]
            phase_residual_all[k, :, :] = residual.reshape((N_sources, N_stations))
        except:
            # Set screen to zero if fit did not work
            logging.debug('Screen fit failed for timeslot {}'.format(k))
            real_fit_white_all[k, :, :] = np.zeros((N_sources, N_stations))
            real_residual_all[k, :, :] = np.ones((N_sources, N_stations))
            imag_fit_white_all[k, :, :] = np.zeros((N_sources, N_stations))
            imag_residual_all[k, :, :] = np.ones((N_sources, N_stations))
            phase_fit_white_all[k, :, :] = np.zeros((N_sources, N_stations))
            phase_residual_all[k, :, :] = np.ones((N_sources, N_stations))

    outQueue.put([real_fit_white_all, real_residual_all,
                  imag_fit_white_all, imag_residual_all,
                  phase_fit_white_all, phase_residual_all,
                  times])


def _fit_tec_screen(station_names, source_names, pp, airmass, rr, weights, times,
    height, order, r_0, beta, outQueue):
    """
    Fits a screen to given TEC values using Karhunen-Lo`eve base vectors

    Parameters
    ----------
    station_names: array
        Array of station names
    source_names: array
        Array of source names
    pp: array
        Array of piercepoint locations
    airmass: array
        Array of airmass values (note: not currently used)
    rr: array
        Array of TEC values to fit screen to
    weights: array
        Array of weights
    times: array
        Array of times
    height: float
        Height of screen (m)
    order: int
        Order of screen (i.e., number of KL base vectors to keep)
    r_0: float
        Scale size of phase fluctuations (m)
    beta: float
        Power-law index for phase structure function (5/3 => pure Kolmogorov
        turbulence)

    """
    import numpy as np
    from pylab import kron, concatenate, pinv, norm, newaxis, find, amin, svd, eye

    logging.info('Fitting screens...')

    # Initialize arrays
    N_stations = len(station_names)
    N_sources = len(source_names)
    N_times = len(times)
    N_piercepoints = N_sources * N_stations
    tec_fit_white_all = np.zeros((N_times, N_sources, N_stations))
    tec_residual_all = np.zeros((N_times, N_sources, N_stations))

    for k in range(N_times):
        D = np.resize(pp[k, :, :], (N_piercepoints, N_piercepoints, 3))
        D = np.transpose(D, (1, 0, 2)) - D
        D2 = np.sum(D**2, axis=2)
        C = -(D2 / r_0**2)**(beta / 2.0) / 2.0
        pinvC = pinv(C, rcond=1e-3)
        U, S, V = svd(C)
        invU = pinv(np.dot(np.transpose(U[:, :order]), np.dot(weights[:, :, k], U[:, :order])), rcond=1e-3)

        # Calculate screen
        rr1 = np.dot(np.transpose(U[:, :order]), np.dot(weights[:, :, k], rr[:, k]))
        tec_fit = np.dot(pinvC, np.dot(U[:, :order], np.dot(invU, rr1)))
        tec_fit_white_all[k, :, :] = tec_fit.reshape((N_sources, N_stations))
        residual = rr - np.dot(C, tec_fit)[:, newaxis]
        tec_residual_all[k, :, :] = residual.reshape((N_sources, N_stations))

    outQueue.put([tec_fit_white_all, tec_residual_all, times])


def run(soltab, outSoltab='tecscreen', height=200.0e3, order=12,
    beta=5.0/3.0, ncpu=0):
    """
    Fits a screen to TEC + scalaraphase values.

    The results of the fit are stored in the soltab parent solset in
    "outSoltab" and the residual phases (actual-screen) are stored in
    "outsoltabresid". These values are the screen phase values per station per
    pierce point per solution interval. The pierce point locations are stored in
    an auxiliary array in the output soltabs.

    Screens can be plotted with the PLOTSCREEN operation.

    Parameters
    ----------
    soltab: solution table
        Soltab containing phase solutions
    outSoltab: str, optional
        Name of output soltab
    height : float, optional
        Height in m of screen
    order : int, optional
        Order of screen (i.e., number of KL base vectors to keep).
    beta: float, optional
        Power-law index for phase structure function (5/3 => pure Kolmogorov
        turbulence)
    ncpu: int, optional
        Number of CPUs to use. If 0, all are used
    niter: int, optional
        Number of iterations to do when determining weights
    nsigma: float, optional
        Number of sigma above which directions are flagged

    """
    import numpy as np
    from numpy import newaxis
    import re
    import os

    # Get screen type
    screen_type = soltab.getType()
    if screen_type not in ['phase', 'tec']:
        logging.error('Screens can only be fit to soltabs of type "phase" or "tec".')
        return 1
    logging.info('Using solution table {0} to calculate {1} screens'.format(soltab.name, screen_type))

    # Load phases
    axis_names = soltab.getAxesNames()
    r = np.array(soltab.val)
    weights = soltab.weight[:]
    if 'freq' in soltab.getAxesNames():
        freqs = soltab.freq[:]
        if len(freqs) > 1:
            logging.error('Screens can only be fit at a single frequency')
            return 1
        freq = freqs[0]

        # remove degenerate freq axis
        freq_ind = soltab.getAxesNames().index('freq')
        r = np.squeeze(r, axis=freq_ind)
        weights = np.squeeze(weights, axis=freq_ind)
        axis_names.pop(freq_ind)

    # fix for missing dir axis
    if not 'dir' in soltab.getAxesNames():
        r = np.array([r])
        weights = np.array([weights])
        dir_ind = len(axis_names)
        source_names = ['POINTING']
    else:
        dir_ind = axis_names.index('dir')
        source_names = soltab.dir[:]

    time_ind = axis_names.index('time')
    ant_ind = axis_names.index('ant')
    r = r.transpose([dir_ind, ant_ind, time_ind])
    weights = weights.transpose([dir_ind, ant_ind, time_ind])
    times = np.array(soltab.time)

    # Collect station and source names and positions and times, making sure
    # that they are ordered correctly.
    solset = soltab.getSolset()
    source_dict = solset.getSou()
    source_positions = []
    for source in source_names:
        source_positions.append(source_dict[source])
    station_names = soltab.ant[:]
    station_dict = solset.getAnt()
    station_positions = []
    for station in station_names:
        station_positions.append(station_dict[station])
    N_sources = len(source_names)
    N_times = len(times)
    N_stations = len(station_names)

    logging.info('Using height = {0} m and order = {1}'.format(height, order))
    if height < 100e3:
        logging.warning("Height is less than 100e3 m. This is likely too low.")

    # Initialize various arrays
    N_piercepoints = N_sources * N_stations
    if screen_type == 'phase':
        real_screen = np.zeros((N_sources, N_stations, N_times))
        real_residual =np.zeros((N_sources, N_stations, N_times))
        imag_screen = np.zeros((N_sources, N_stations, N_times))
        imag_residual = np.zeros((N_sources, N_stations, N_times))
    screen = np.zeros((N_sources, N_stations, N_times))
    residual = np.zeros((N_sources, N_stations, N_times))
    val_amp = 1.0
    r_0 = 100.0 # shouldn't matter what we choose
    rr = np.reshape(r, [N_piercepoints, N_times])

    # Find pierce points and airmass values for given screen height
    pp, airmass, midRA, midDec = _calculate_piercepoints(np.array(station_positions),
        np.array(source_positions), np.array(times), height)

    # Fit the screens
    station_weights = np.reshape(weights, [N_piercepoints, N_times])
    if screen_type == 'phase':
        mpm = multiprocManager(ncpu, _fit_phase_screen)
        for tindx, t in enumerate(times):
            w = np.diag(station_weights[:, tindx])[:, :, newaxis]
            mpm.put([station_names, source_names, pp[tindx, newaxis, :, :],
                airmass[tindx, newaxis, :], rr[:, tindx, newaxis], w,
                [t], height, order, r_0, beta])
        mpm.wait()
        for (real_scr, real_res, imag_scr, imag_res, phase_scr, phase_res, t) in mpm.get():
            i = times.tolist().index(t[0])
            real_screen[:, :, i] = real_scr[0, :, :]
            real_residual[:, :, i] = real_res[0, :, :]
            imag_screen[:, :, i] = imag_scr[0, :, :]
            imag_residual[:, :, i] = imag_res[0, :, :]
            screen[:, :, i] = phase_scr[0, :, :]
            residual[:, :, i] = phase_res[0, :, :]
    elif screen_type == 'tec':
        mpm = multiprocManager(ncpu, _fit_tec_screen)
        for tindx, t in enumerate(times):
            w = np.diag(station_weights[:, tindx])[:, :, newaxis]
            mpm.put([station_names, source_names, pp[tindx, newaxis, :, :],
                airmass[tindx, newaxis, :], rr[:, tindx, newaxis], w,
                [t], height, order, r_0, beta])
        mpm.wait()
        for (scr, res, t) in mpm.get():
            i = times.tolist().index(t[0])
            screen[:, :, i] = scr[0, :, :]
            residual[:, :, i] = res[0, :, :]
    weights = np.reshape(station_weights, (N_sources, N_stations, N_times))

    # Write the results to the output solset
    dirs_out = source_names
    times_out = times
    ants_out = station_names

    # Store screen values
    weights = weights.transpose([0, 2, 1]) # order is now [dir, time, ant]
    vals = screen.transpose([0, 2, 1])
    screen_st = solset.makeSoltab('{}screen'.format(screen_type), outSoltab,
        axesNames=['dir', 'time', 'ant'], axesVals=[dirs_out, times_out,
        ants_out], vals=vals, weights=weights)
    vals = residual.transpose([0, 2, 1])
    resscreen_st = solset.makeSoltab('{}screenresid'.format(screen_type), outSoltab+'resid',
        axesNames=['dir', 'time', 'ant'], axesVals=[dirs_out, times_out,
        ants_out], vals=vals, weights=weights)

    # Store beta, r_0, height, and order as attributes of the screen soltabs
    screen_st.obj._v_attrs['beta'] = beta
    screen_st.obj._v_attrs['r_0'] = r_0
    screen_st.obj._v_attrs['height'] = height
    screen_st.obj._v_attrs['order'] = order
    if 'freq' in soltab.getAxesNames():
        screen_st.obj._v_attrs['freq'] = freq

    # Store piercepoint table. Note that it does not conform to the axis
    # shapes, so we cannot use makeSoltab()
    solset.obj._v_file.create_array('/'+solset.name+'/'+screen_st.obj._v_name,
        'piercepoint', obj=pp)

    screen_st.addHistory('CREATE (by DIRECTIONSCREEN operation)')
    resscreen_st.addHistory('CREATE (by DIRECTIONSCREEN operation)')

    return 0



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This tool is used to import killms solutions into a H5parm format.
"""
# Authors:
# Francesco de Gasperin
_author = "Francesco de Gasperin (astro@voo.it)"

import sys, os, glob, pickle
import numpy as np
from losoto import _version
from losoto import _logging
from losoto import h5parm as h5parm_mod
    
if __name__=='__main__':
    # Options
    import optparse
    opt = optparse.OptionParser(usage='%prog [-v] <H5parm> <killmsfile> \n'\
                    +_author, version='%prog '+_version.__version__)
    opt.add_option('-V', '--verbose', help='Go Vebose! (default=False)', action='store_true', default=False)
    opt.add_option('-s', '--solset', help='Solution-set name (default=sol###)', type='string', default=None)
    opt.add_option('-c', '--complevel', help='Compression level from 0 (no compression, fast) to 9 (max compression, slow) (default=5)', type='int', default='5')
    opt.add_option('--nofulljones', help='Export the solutions as just XX/YY instead of XX,XY,YX,YY.', action='store_false', dest='fulljones', default=True)
    (options, args) = opt.parse_args()

    # Check options
    if len(args) != 2:
        opt.print_help()
        sys.exit()

    # log
    logger = _logging.Logger('info')
    logging = _logging.logger
    if options.verbose: logger.set_level("debug")

    inputFile = args[1]
    logging.info("KILLMS filenames = "+str(inputFile))
    h5parmFile = args[0]
    logging.info("H5parm filename = "+h5parmFile)
    
    # Common options
    complevel = options.complevel
    solsetName = options.solset

    SolsDico = np.load(inputFile)
    Sols = SolsDico["Sols"]
    Sols = Sols.view(np.recarray)

    # build direction subtable
    ClusterCat = SolsDico["ClusterCat"]
    dirCoords = []; dirNames = []
    for i, c in enumerate(ClusterCat):
        dirCoords.append([c[1], c[2]])
        dirNames.append('Dir%02i' % i)

    # build antenna subtable
    stationNames = SolsDico["StationNames"]
    antPos = []; antNames = []
    for i, a in enumerate(stationNames):
        antPos.append([0,0,0])
        antNames.append(a)

    #print SolsDico.keys()
    #print Sols.dtype.names
    times = (Sols["t0"]+Sols["t1"])/2.
    freqs = (SolsDico['FreqDomains'][:,0]+SolsDico['FreqDomains'][:,1])/2.

    # fix some kMS weirdness with first and last timeslot values
    #otherwise they are off the time solution grid
    mediandiff = np.median(Sols["t1"] - Sols["t0"])
    times[0]  = times[1] - mediandiff
    times[-1] = times[-2] + mediandiff
    
    
    # construct solution arrays
    if options.fulljones:
        pols = ['XX','XY','YX','YY']
        tt, tf, ta, td, _, _ = Sols['G'].shape
        vals_amp = np.zeros(shape=(tt,tf,ta,td,4))
        vals_ph = np.zeros(shape=(tt,tf,ta,td,4))
        vals_amp[...,0] = np.abs(Sols['G'][...,0,0])
        vals_amp[...,1] = np.abs(Sols['G'][...,0,1])
        vals_amp[...,2] = np.abs(Sols['G'][...,1,0])
        vals_amp[...,3] = np.abs(Sols['G'][...,1,1])
        vals_ph[...,0] = np.angle(Sols['G'][...,0,0])
        vals_ph[...,1] = np.angle(Sols['G'][...,0,1])
        vals_ph[...,2] = np.angle(Sols['G'][...,1,0])
        vals_ph[...,3] = np.angle(Sols['G'][...,1,1])
    elif not options.fulljones:
        pols = ['XX','YY']
        tt, tf, ta, td, _, _ = Sols['G'].shape
        vals_amp = np.zeros(shape=(tt,tf,ta,td,2))
        vals_ph = np.zeros(shape=(tt,tf,ta,td,2))
        vals_amp[...,0] = np.abs(Sols['G'][...,0,0])
        vals_amp[...,1] = np.abs(Sols['G'][...,1,1])
        vals_ph[...,0] = np.angle(Sols['G'][...,0,0])
        vals_ph[...,1] = np.angle(Sols['G'][...,1,1])
        
    #print vals_amp.shape
    #print len(pols), len(dirNames), len(antNames), len(freqs), len(times)

    weights = np.ones(shape=vals_amp.shape)

    is_tec = 'SolsTEC' in list(SolsDico.keys())
    if is_tec:
        # construct TEC array 
        vals_tec = np.zeros(shape=(td,ta,tt))
        vals_tec = SolsDico['SolsTEC'].T
        vals_csp = np.zeros(shape=(td,ta,tt))
        vals_csp = SolsDico['SolsCPhase'].T
        print((vals_tec.shape))
        weights_tec = np.ones(shape=vals_tec.shape)

    # write to h5pram
    h5parm = h5parm_mod.h5parm(h5parmFile, readonly = False, complevel = complevel)
    solset = h5parm.makeSolset(solsetName)

    if is_tec:
        solset.makeSoltab('tec', axesNames=['ant','dir','time'], \
                axesVals=[antNames,dirNames,times], vals=vals_tec, weights=weights_tec)
        solset.makeSoltab('phase', 'offset', axesNames=['ant','dir','time'], \
                axesVals=[antNames,dirNames,times], vals=vals_csp, weights=weights_tec)
    else:
        solset.makeSoltab('amplitude', axesNames=['time','freq','ant','dir','pol'], \
                axesVals=[times,freqs,antNames,dirNames,pols], vals=vals_amp, weights=weights)
        solset.makeSoltab('phase', axesNames=['time','freq','ant','dir','pol'], \
                axesVals=[times,freqs,antNames,dirNames,pols], vals=vals_ph, weights=weights)

    # fill source table
    sourceTable = solset.obj._f_get_child('source')
    sourceTable.append(list(zip(*(dirNames,dirCoords))))

    # fill antenna table
    antennaTable = solset.obj._f_get_child('antenna')
    antennaTable.append(list(zip(*(antNames,antPos))))
    print(h5parm)









#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This is an example operation for LoSoTo

from losoto.lib_operations import *
from losoto._logging import logger as logging

logging.debug('Loading EXAMPLE module.')

# this funct is called by losoto to set parameters and call the real run()
def _run_parser(soltab, parser, step):
    opt1 = parser.getfloat( step, 'opt1') # no default
    opt2 = parser.getarrayfloat( step, 'opt3', [1., 2., 3.])
    opt3 = parser.getint( step, 'opt2', 0 )
    
    parser.checkSpelling( step, soltab, ['op1', 'opt2', 'opt3'])
    return run(soltab, opt1, opt2, opt3)

# this funct can be called by python directly
# parameters that are non optional require the default value equal to the one defined for the parset above
def run( soltab, opt1, opt2 = [1., 2., 3.], opt3 = 0 ):
    """
    Generic unspecified step for easy expansion.

    Parameters
    ----------
    opt1 : float
        Is a mandatory parameter.

    opt2 : list of float, optional
        Is optional, by default [1.,2.,3.]

    opt2 : int, optional
        Is optional, by default 0.
    """

    # load specific libs
    import numpy as np
 
    # initial logging
    logging.info("Working on soltab: "+soltab.name)
    
    # check input
    # ...

    axisNames = soltab.getAxesNames()
    logging.info("Axis names are: "+str(axisNames))
 
    solType = soltab.getType()
    logging.info("Soltab type is: "+solType)
 
    soltab.setSelection(ant=soltab.getAxisValues('ant')[0])
    logging.info("Selection is: "+str(soltab.selection))
 
    # find axis values
    logging.info("Antennas (no selection) are: "+str(soltab.getAxisValues('ant', ignoreSelection=True)))
    logging.info("Antennas (with selection) are: "+str(soltab.getAxisValues('ant')))
    # but one can also use (selection is active here!)
    logging.info("Antennas (other method) are: "+str(soltab.ant))
    logging.info("Frequencies are: "+str(soltab.freq))
    logging.info("Directions are: "+str(soltab.dir))
    logging.info("Polarizations are: "+str(soltab.pol))
    # try to access a non-existent axis
    soltab.getAxisValues('nonexistantaxis')
 
    # now get all values given this selection
    logging.info("Get data using soltab.val")
    val = soltab.val
    logging.debug('shape of val: '+str(soltab.val.shape))
    logging.info("$ val is "+str(val[0,0,0,0,100]))
    weight = soltab.weight
    time = soltab.time
    thisTime = soltab.time[100]
 
    # another way to get the data is using the getValues()
    logging.info("Get data using getValues()")
    grid, axes = soltab.getValues()
    # axis names
    logging.info("Axes: "+str(soltab.getAxesNames()))
    # axis shape
    print(axes)
    print([soltab.getAxisLen(axis) for axis in axes]) # not ordered, is a dict!
    # data array shape (same of axis shape)
    logging.info("Shape of values: "+str(grid.shape))
    #logging.info("$ val is "+str(grid[0,0,0,0,100]))
 
    # reset selection
    soltab.setSelection()
    logging.info('Reset selection to \'\'')
    logging.info("Antennas are: "+str(soltab.ant))
    logging.info("Frequencies are: "+str(soltab.freq))
    logging.info("Directions are: "+str(soltab.dir))
    logging.info("Polarizations are: "+str(soltab.pol))
 
    # finally the getValuesIter allaws to iterate across all possible combinations of a set of axes
    logging.info('Iteration on time/freq')
    for vals, coord, selection in soltab.getValuesIter(returnAxes=['time','freq']):
        # writing back the solutions
        soltab.setValues(vals, selection)
    logging.info('Iteration on time')
    for vals, coord, selection in soltab.getValuesIter(returnAxes=['time']):
        # writing back the solutions
        soltab.setValues(vals, selection)   
    logging.info('Iteration on dir after selection to 1 dir')
    soltab.setSelection(dir='pointing') 
    for vals, coord, selection in t.getValuesIter(returnAxes=['dir']):
        # writing back the solutions
        soltab.setValues(vals, selection)
  
    return 0 # if everything went fine, otherwise 1
 
 



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# This tool is used to merge 2 H5parms.
# Each h5parm solset will be copied into the last one.

_author = "Francesco de Gasperin (astro@voo.it)"

import sys, os, glob
import numpy as np
import pyrap.tables as pt
from losoto import _version
from losoto import _logging
import losoto.h5parm

if __name__=='__main__':
    # Options
    import optparse
    opt = optparse.OptionParser(usage='%prog [-v] <H5parm:solset> <H5parm:solset> \n'\
                            +_author, version='%prog '+_version.__version__)
    opt.add_option('-V', '--verbose', help='Go VERBOSE! (default=False)', action='store_true', default=False)
    (options, args) = opt.parse_args()

    # Check options
    if len(args) != 2:
        opt.print_help()
        sys.exit()

    # log
    logger = _logging.Logger('info')
    logging = _logging.logger
    if options.verbose: logger.set_level("debug")

    h5parmFrom = args[0]
    h5parmTo = args[1]

    logging.info("H5parm origin = "+h5parmFrom)
    logging.info("H5parm destination = "+h5parmTo)

    # scompose input values
    h5parmFromFile, solsetFrom = h5parmFrom.split(':')
    h5parmToFile, solsetTo = h5parmTo.split(':')

    if not (os.path.isfile(h5parmFromFile) and os.path.isfile(h5parmToFile)):
        logging.critical("Missing H5parm file.")
        sys.exit(1)

    # retrieve table
    hf = losoto.h5parm.h5parm(h5parmFromFile)
    ssF = hf.getSolset(solset=solsetFrom)

    # write table
    ht = losoto.h5parm.h5parm(h5parmToFile, readonly=False)
    # check if the solset exists
    if solsetTo in ht.getSolsetNames():
        logging.critical('Destination solset already exists, quitting.')
        sys.exit(1)
    ssT = ht.makeSolset(solsetName = solsetTo, addTables=False)
    # write the soltabs
    ssF.obj._f_copy_children(ssT.obj, recursive=True)

    del hf

    # Add entry to history
    soltabs = ht.getSolset(solsetTo).getSoltabs()
    for st in soltabs:
        st.addHistory('Copied (from {0}:{1})'.format(h5parmFromFile, solsetFrom))
    del ht

    logging.info("Done.")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from losoto.lib_operations import *
from losoto._logging import logger as logging

logging.debug('Loading REPLICATEONAXIS module.')

def _run_parser(soltab, parser, step):
    axisReplicate = parser.getstr( step, 'axisReplicate' ) # no default
    fromCell = parser.getstr( step, 'fromCell', '' )
    updateWeights = parser.getbool( step, 'updateWeights', True )

    parser.checkSpelling( step, soltab, ['axisReplicate', 'fromCell', 'updateWeights'])
    return run(soltab, axisReplicate, fromCell, updateWeights)

def run( soltab, axisReplicate, fromCell, updateWeights=True):
    """
    Replace the values along a certain axis taking them from one specific axic cell

    Parameters
    ----------
    axisReplicate : str
        Axis along which replicate the values.

    fromCell : str
        A cell value in axisReplicate from which to copy the data values. If it is the string
        "first"/"last" then uses the first/last element of the axis.

    updateWeights : bool
        If False then weights are untoched, if True they are replicated like data. Default: True.
    """
    import numpy as np

    if not axisReplicate in soltab.getAxesNames():
        logging.error('Cannot find axis %s.' % axisReplicate)
        return 1

    if fromCell == 'first':
        fromCell = soltab.getAxisValues(axisReplicate)[0]
    elif fromCell == 'last':
        fromCell = soltab.getAxisValues(axisReplicate)[-1]

    axisType = type(soltab.getAxisValues(axisReplicate)[0])
    try:
        fromCell = np.array([fromCell]).astype(axisType)[0]
    except:
        logging.error('Cannot convert to type %s the value in fromCell: %s.' % (str(axisType),fromCell))
        return 1

    if not fromCell in soltab.getAxisValues(axisReplicate):
        logging.error('Cannot find %s in %s.' % (fromCell, axisReplicate))
        return 1

    logging.info("Replicate axis on soltab: "+soltab.name)

    # get the cell to replicate
    axisReplicateLen = soltab.getAxisLen(axisReplicate, ignoreSelection=False) # keep selection into account
    old_selection = soltab.selection

    # get slice with 1 value to replicate
    soltab.setSelection(**{axisReplicate:fromCell})
    vals = soltab.getValues(retAxesVals=False)
    if updateWeights:
        weights = soltab.getValues(retAxesVals=False, weight=True)

    cellPos = list(soltab.getAxisValues(axisReplicate)).index(fromCell)
    axisReplicatePos = soltab.getAxesNames().index(axisReplicate)

    # expand on the right axis
    vals = np.repeat(vals, repeats=axisReplicateLen, axis=axisReplicatePos)

    # write back
    soltab.selection = old_selection
    soltab.setValues(vals)
    if updateWeights:
        soltab.setValues(weights, weight=True)

    soltab.addHistory('REPLICATEONAXIS (over axis %s from cell %s)' % (axisReplicate, str(fromCell)))
    return 0



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from losoto.lib_operations import *
from losoto._logging import logger as logging

logging.debug('Loading INTERPOLATE module.')

def _run_parser(soltab, parser, step):
    outSoltab = parser.getstr( step, 'outSoltab') # no default
    axisToRegrid = parser.getstr( step, 'axisToRegrid') # no default
    newDelta = parser.getstr( step, 'newDelta') # no default
    delta = parser.getstr( step, 'delta', '')
    maxFlaggedWidth = parser.getint( step, 'maxFlaggedWidth', 0)
    log = parser.getbool( step, 'log', False )

    parser.checkSpelling( step, soltab, ['outSoltab', 'axisToRegrid', 'newDelta', 'delta', 'maxFlaggedWidth', 'log'])
    return run( soltab, outSoltab, axisToRegrid, newDelta, delta, maxFlaggedWidth, log)


def _regrid_axis(vals, delta, newdelta):
    """
    Regrids an axis

    Parameters
    ----------
    vals : array
        Array of axis values to regrid

    delta : float
        Fundamental sampling width of input values

    newdelta : float
        Desired sampling width of regridded values

    Returns
    -------
    vals_new : array
        Array of new values
    """
    offset = newdelta / 2.0 - 0.5 * delta
    freqmin = np.min(vals) + offset
    freqmax = np.max(vals) + offset
    vals_new  = np.arange(freqmin, freqmax + newdelta / 2.0, newdelta)

    return vals_new


def _convert_strval(val):
    """
    Converts a string value with units to a float

    Parameters
    ----------
    val : str or float
        Value to convert. Units can be seconds ("s") or "Hz". E.g., "100kHz" or "10s"

    Returns
    -------
    val : float
        Value in fundamental units (s or Hz)
    """
    if type(val) is str:
        letters = [1 for s in val[::-1] if s.isalpha()]
        indx = len(val) - sum(letters)
        unit = val[indx:]
        if unit.strip().lower() == 'hz' or unit.strip().lower() == 's':
            conversion = 1.0
        elif unit.strip().lower() == 'khz' or unit.strip().lower() == 'ks':
            conversion = 1e3
        elif unit.strip().lower() == 'mhz' or unit.strip().lower() == 'ms':
            conversion = 1e6
        else:
            logging.error("The unit on delta was not understood.")
            return 1
        val = float(val[:indx]) * conversion
    else:
        val = val

    return val


def run( soltab, outsoltab, axisToRegrid, newdelta, delta='', maxFlaggedWidth=0, log=False):
    """
    This operation for LoSoTo implements regridding and linear interpolation of data for an axis.
    WEIGHT: compliant

    Parameters
    ----------
    outsoltab: str
        Name of output soltab

    axisToRegrid : str
        Name of the axis for which regridding/interpolation will be done

    newdelta : float or str
        Fundamental width between samples after regridding. E.g., "100kHz" or "10s"

    delta : float or str, optional
        Fundamental width between samples in axisToRegrid. E.g., "100kHz" or "10s". If "",
        it is calculated from the axisToRegrid values

    maxFlaggedWidth : int, optional
        Maximum allowable width in number of samples (after regridding) above which
        interpolated values are flagged (e.g., maxFlaggedWidth = 5 would allow gaps of
        5 samples or less to be interpolated across but gaps of 6 or more would be
        flagged)

    log : bool, optional
        Interpolation is done in log10 space, by default False
    """
    import scipy.ndimage as nd

    # Check inputs
    if axisToRegrid not in soltab.getAxesNames():
        logging.error('Axis \"'+axisToRegrid+'\" not found.')
        return 1
    if axisToRegrid not in ['freq', 'time']:
        logging.error('Axis \"'+axisToRegrid+'\" must be either time or freq.')
        return 1
    newdelta = _convert_strval(newdelta)
    if delta == "":
        deltas = soltab.getAxisValues(axisToRegrid)[1:] - soltab.getAxisValues(axisToRegrid)[:-1]
        delta = np.min(deltas)
        logging.info('Using {} for delta'.format(delta))
    else:
        delta = _convert_strval(delta)
    if soltab.getType() == 'amplitude' and not log:
        logging.warning('Amplitude solution tab detected and log=False. Amplitude solution tables should be treated in log space.')

    # Regrid axis
    axisind = soltab.getAxesNames().index(axisToRegrid)
    orig_axisvals = soltab.getAxisValues(axisToRegrid)
    new_axisvals = _regrid_axis(orig_axisvals, delta, newdelta)
    orig_shape = soltab.val.shape
    new_shape = list(orig_shape)
    new_shape[axisind] = len(new_axisvals)
    new_vals = np.zeros(new_shape, dtype='float')
    new_weights = np.zeros(new_shape, dtype='float')

    for vals, weights, coord, selection in soltab.getValuesIter(returnAxes=[axisToRegrid], weight=True):
        flagged = np.logical_or(np.equal(weights, 0.0), np.isnan(vals))
        weights[flagged] = 0.0
        unflagged = np.not_equal(weights, 0.0)
        if np.sum(unflagged) > 2:
            # If there are at least two unflagged points, interpolate with mask
            if log:
                vals = np.log10(vals)
            new_vals[selection] = np.interp(new_axisvals, orig_axisvals[unflagged],
                                            vals[unflagged], left=np.nan, right=np.nan)

            # For the weights, interpolate without the mask
            new_weights[selection] = np.round(np.interp(new_axisvals, orig_axisvals, weights,
                                                        left=np.nan, right=np.nan))

        # Check for flagged gaps
        if maxFlaggedWidth > 1:
            inv_weights = new_weights[selection].astype(bool).squeeze()
            rank = len(inv_weights.shape)
            connectivity = nd.generate_binary_structure(rank, rank)
            mask_labels, count = nd.label(~inv_weights, connectivity)
            for i in range(count):
                ind = np.where(mask_labels == i+1)
                gapsize = len(ind[0])
                if gapsize <= maxFlaggedWidth:
                    # Unflag narrow gaps
                    selection[axisind] = ind[0]
                    new_weights[selection] = 1.0

    # Write new soltab
    solset = soltab.getSolset()
    axesVals = []
    for axisName in soltab.getAxesNames():
        if axisName == axisToRegrid:
            axesVals.append(new_axisvals)
        else:
            axesVals.append(soltab.getAxisValues(axisName))
    if log:
        new_vals = 10**new_vals
    s = solset.makeSoltab(soltab.getType(), outsoltab,
        axesNames=soltab.getAxesNames(), axesVals=axesVals, vals=new_vals, weights=new_weights)
    s.addHistory('CREATE by INTERPOLATE operation from '+soltab.name+'.')

    return 0



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from losoto.lib_operations import *
from losoto._logging import logger as logging

logging.debug('Loading PREFACTOR_BANDPASS module.')

def _run_parser(soltab, parser, step):
    chanWidth = parser.getstr( step, 'chanWidth', '')
    outSoltabName = parser.getstr( step, 'outSoltabName', 'bandpass' )
    BadSBList = parser.getstr( step, 'BadSBList', '')
    interpolate = parser.getbool( step, 'interpolate', True )
    removeTimeAxis = parser.getbool( step, 'removeTimeAxis', True )
    autoFlag = parser.getbool( step, 'autoFlag', False )
    nSigma = parser.getfloat( step, 'nSigma', 5.0 )
    maxFlaggedFraction = parser.getfloat( step, 'maxFlaggedFraction', 0.5 )
    maxStddev = parser.getfloat( step, 'maxStddev', 0.01 )
    ncpu = parser.getint( '_global', 'ncpu', 0 )

    parser.checkSpelling( step, soltab, ['chanWidth', 'outSoltabName', 'BadSBList', 'interpolate', 'removeTimeAxis',\
               'autoFlag', 'nSigma', 'maxFlaggedFraction', 'maxStddev'])
    return run(soltab, chanWidth, outSoltabName, BadSBList, interpolate, removeTimeAxis,
               autoFlag, nSigma, maxFlaggedFraction, maxStddev, ncpu)

def _savitzky_golay(y, window_size, order, deriv=0, rate=1):
    """Smooth (and optionally differentiate) data with a Savitzky-Golay filter.
    The Savitzky-Golay filter removes high frequency noise from data.
    It has the advantage of preserving the original shape and
    features of the signal better than other types of filtering
    approaches, such as moving averages techniques.

    Parameters
    ----------
    y : array_like, shape (N,)
        the values of the time history of the signal.
    window_size : int
        the length of the window. Must be an odd integer number.
    order : int
        the order of the polynomial used in the filtering.
        Must be less then `window_size` - 1.
    deriv: int
        the order of the derivative to compute (default = 0 means only smoothing)

    Returns
    -------
    ys : ndarray, shape (N)
        the smoothed signal (or it's n-th derivative).

    Notes
    -----
    The Savitzky-Golay is a type of low-pass filter, particularly
    suited for smoothing noisy data. The main idea behind this
    approach is to make for each point a least-square fit with a
    polynomial of high order over a odd-sized window centered at
    the point.

    Examples
    --------
    t = np.linspace(-4, 4, 500)
    y = np.exp( -t**2 ) + np.random.normal(0, 0.05, t.shape)
    ysg = savitzky_golay(y, window_size=31, order=4)
    import matplotlib.pyplot as plt
    plt.plot(t, y, label='Noisy signal')
    plt.plot(t, np.exp(-t**2), 'k', lw=1.5, label='Original signal')
    plt.plot(t, ysg, 'r', label='Filtered signal')
    plt.legend()
    plt.show()

    References
    ----------
    .. [1] A. Savitzky, M. J. E. Golay, Smoothing and Differentiation of
       Data by Simplified Least Squares Procedures. Analytical
       Chemistry, 1964, 36 (8), pp 1627-1639.
    .. [2] Numerical Recipes 3rd Edition: The Art of Scientific Computing
       W.H. Press, S.A. Teukolsky, W.T. Vetterling, B.P. Flannery
       Cambridge University Press ISBN-13: 9780521880688
    """
    import numpy as np
    from math import factorial

    try:
        window_size = np.abs(np.int(window_size))
        order = np.abs(np.int(order))
    except ValueError as msg:
        raise ValueError("window_size and order have to be of type int")
    if window_size % 2 != 1 or window_size < 1:
        raise TypeError("window_size size must be a positive odd number")
    if window_size < order + 2:
        raise TypeError("window_size is too small for the polynomials order")
    order_range = list(range(order+1))
    half_window = (window_size -1) // 2
    # precompute coefficients
    b = np.mat([[k**i for i in order_range] for k in range(-half_window, half_window+1)])
    m = np.linalg.pinv(b).A[deriv] * rate**deriv * factorial(deriv)
    # pad the signal at the extremes with
    # values taken from the signal itself
    firstvals = y[0] - np.abs( y[1:half_window+1][::-1] - y[0] )
    lastvals = y[-1] + np.abs(y[-half_window-1:-1][::-1] - y[-1])
    y = np.concatenate((firstvals, y, lastvals))
    return np.convolve( m[::-1], y, mode='valid')


def _B(x, k, i, t, extrap, invert):
    if k == 0:
        if extrap:
            if invert:
                return -1.0
            else:
                return 1.0
        else:
            return 1.0 if t[i] <= x < t[i+1] else 0.0
    if t[i+k] == t[i]:
       c1 = 0.0
    else:
       c1 = (x - t[i])/(t[i+k] - t[i]) * _B(x, k-1, i, t, extrap, invert)
    if t[i+k+1] == t[i+1]:
       c2 = 0.0
    else:
       c2 = (t[i+k+1] - x)/(t[i+k+1] - t[i+1]) * _B(x, k-1, i+1, t, extrap, invert)
    return c1 + c2


def _bspline(x, t, c, k):
    n = len(t) - k - 1
    assert (n >= k+1) and (len(c) >= n)
    invert = False
    extrap = [False] * n
    if x >= t[n]:
        extrap[-1] = True
    elif x < t[k]:
        extrap[0] = True
        invert = False
    return sum(c[i] * _B(x, k, i, t, e, invert) for i, e in zip(list(range(n)), extrap))


def _bandpass_LBA(freq, c1, c2, c3, c4, c5, c6, c7, c8, c9, c10, c11, c12, c13):
    """
    Defines the functional form of the LBA bandpass in terms of splines of degree 3

    The spline fit was done using LSQUnivariateSpline() on the median bandpass between
    30 MHz and 78 MHz. The knots were set by hand to acheive a good fit with a
    minimum number of parameters.

    Parameters
    ----------
    freq : array
        Array of frequencies
    c1-c9: float
        Spline coefficients

    Returns
    -------
    bandpass : list
        List of bandpass values as function of frequency
    """
    knots = np.array([30003357.0, 30003357.0, 30003357.0, 30003357.0, 40000000.0,
                      50000000.0, 55000000.0, 56000000.0, 60000000.0, 62000000.0,
                      63000000.0, 64000000.0, 70000000.0, 77610779.0, 77610779.0,
                      77610779.0, 77610779.0])
    coeffs = np.array([c1, c2, c3, c4, c5, c6, c7, c8, c9, c10, c11, c12, c13])
    return [_bspline(f, knots, coeffs, 3) for f in freq]


def _bandpass_HBA_low(freq, c1, c2, c3, c4, c5, c6, c7, c8, c9, c10):
    """
    Defines the functional form of the HBA-low bandpass in terms of splines of degree 3

    The spline fit was done using LSQUnivariateSpline() on the median bandpass between
    120 MHz and 188 MHz. The knots were set by hand to acheive a good fit with a
    minimum number of parameters.

    Parameters
    ----------
    freq : array
        Array of frequencies
    c1-c9: float
        Spline coefficients

    Returns
    -------
    bandpass : list
        List of bandpass values as function of frequency
    """
    knots = np.array([1.15e+08, 1.15e+08, 1.15e+08, 1.15e+08,
                      1.30e+08, 1.38e+08, 1.48e+08, 1.60e+08,
                      1.68e+08, 1.78e+08, 1.90e+08, 1.90e+08,
                      1.9e+08, 1.9e+08])
    coeffs = np.array([c1, c2, c3, c4, c5, c6, c7, c8, c9, c10])
    return [_bspline(f, knots, coeffs, 3) for f in freq]


def _fit_bandpass(freq, logamp, sigma, band, do_fit=True):
    """
    Fits amplitudes with one of the bandpass functions

    The initial coefficients were determined from a LSQUnivariateSpline() fit on the
    median bandpass of the appropriate band. The allowable fitting ranges were set by
    hand through testing on a number of observations (to allow the bandpass function to
    adjust for the differences between stations but not to fit to RFI, etc.).

    Parameters
    ----------
    freq : array
        Array of frequencies
    amps : array
        Array of log10(amplitudes)
    sigma : array
        Array of sigma (1/weights**2)
    band : str
        Band name ('hba_low', etc.)
    do_fit : bool, optional
        If True, the fitting is done. If False, the unmodified model bandpass is returned

    Returns
    -------
    fit_parms, bandpass : list, list
        List of best-fit parameters, List of bandpass values as function of frequency
    """
    from scipy.optimize import curve_fit

    if band.lower() == 'hba_low':
        bandpass_function = _bandpass_HBA_low
        init_coeffs = np.array([-0.01460369, 0.05062699, 0.02827004, 0.03738518,
                                -0.05729109, 0.02303295, -0.03550487, -0.0803113,
                                -0.2394929, -0.358301])
        bounds_deltas_lower = [0.06, 0.05, 0.04, 0.04, 0.04, 0.04, 0.1, 0.1, 0.2, 0.5]
        bounds_deltas_upper = [0.06, 0.1, 0.1, 0.1, 0.04, 0.04, 0.04, 0.04, 0.05, 0.06]
    elif band.lower() == 'lba':
        bandpass_function = _bandpass_LBA
        init_coeffs = np.array([-0.22654016, -0.1950495, -0.07763014, 0.10002095,
                                0.32797671, 0.46900048, 0.47155583, 0.31945897,
                                0.29072278, 0.08064795, -0.15761538, -0.36020451,
                                -0.51163338])
        bounds_deltas_lower = [0.25, 0.2, 0.05, 0.05, 0.05, 0.05, 0.08, 0.05, 0.08, 0.15,
                               0.15, 0.15, 0.15]
        bounds_deltas_upper = [0.25, 0.2, 0.05, 0.05, 0.05, 0.05, 0.08, 0.05, 0.08, 0.15,
                               0.15, 0.15, 0.15]
    else:
        print(('The "{}" band is not supported'.format(band)))
        sys.exit(1)


    if do_fit:
        lower = [c - b for c, b in zip(init_coeffs, bounds_deltas_lower)]
        upper = [c + b for c, b in zip(init_coeffs, bounds_deltas_upper)]
        param_bounds = (lower, upper)
        try:
            popt, pcov = curve_fit(bandpass_function, freq, logamp, sigma=sigma,
                                   bounds=param_bounds, method='dogbox')
            return popt, bandpass_function(freq, *tuple(popt))
        except RuntimeError:
            logging.error('Fitting failed.' )
            return None, bandpass_function(freq, *tuple(init_coeffs))
    else:
        return None, bandpass_function(freq, *tuple(init_coeffs))


def _flag_amplitudes(freqs, amps, weights, nSigma, maxFlaggedFraction, maxStddev,
                     plot, s, outQueue):
    """
    Flags bad amplitude solutions relative to median bandpass (in log space) by setting
    the corresponding weights to 0.0

    Note: A median over the time axis is done before flagging, so the flags are not time-
    dependent

    Parameters
    ----------
    freqs : array
        Array of frequencies
    amps : array
        Array of amplitudes as [time, ant, freq, pol]
    weights : array
        Array of weights as [time, ant, freq, pol]
    nSigma : float
        Number of sigma for flagging. Amplitudes outside of nSigma*stddev are flagged
    maxFlaggedFraction : float
        Maximum allowable fraction of flagged frequencies. Stations with higher fractions
        will be completely flagged
    maxStddev : float
        Maximum allowable standard deviation
    plot : bool
        If True, the bandpass with flags and best-fit line is plotted for each station
    s : int
        Station index

    Returns
    -------
    indx, weights : int, array
        Station index, modified weights array
    """
    # Determine which band we're in
    if np.median(freqs) < 180e6 and np.median(freqs) > 110e6:
        band = 'hba_low'
        median_min = 50.0
        median_max = 225.0
    elif np.median(freqs) < 90e6:
        band = 'lba'
        median_min = 50.0
        median_max = 200.0
    else:
        print(('The median frequency of {} Hz is outside of any know band'.format(np.median(freqs))))
        sys.exit(1)

    # Skip fully flagged stations
    if np.all(weights == 0.0):
        outQueue.put([s, weights])
        return

    # Build arrays for fitting
    flagged = np.where(np.logical_or(weights == 0.0, amps == 0.0))
    amps_flagged = amps.copy()
    amps_flagged[flagged] = np.nan
    sigma = weights.copy()
    sigma[flagged] = 1.0
    sigma = np.sqrt(1.0 / sigma)
    sigma[flagged] = 1e8

    # Iterate over polarizations
    npols = amps.shape[2] # number of polarizations
    for pol in range(npols):
        # take median over time and divide out the median offset
        with np.warnings.catch_warnings():
            np.warnings.filterwarnings('ignore', r'All-NaN (slice|axis) encountered')
            amps_div = np.nanmedian(amps_flagged[:, :, pol], axis=0)
            median_val = np.nanmedian(amps_div)
        amps_div /= median_val
        sigma_div = np.median(sigma[:, :, pol], axis=0)
        median_flagged = np.where(np.isnan(amps_div))
        amps_div[median_flagged] = 1.0
        sigma_div[median_flagged] = 1e8
        median_flagged = np.where(amps_div <= 0.0)
        amps_div[median_flagged] = 1.0
        sigma_div[median_flagged] = 1e8
        sigma_orig = sigma_div.copy()

        # Before doing the fitting, flag any solutions that deviate from the model bandpass by
        # a large factor to avoid biasing the first fit
        _, bp_sp = _fit_bandpass(freqs, np.log10(amps_div), sigma_div, band, do_fit=False)
        bad = np.where(np.abs(bp_sp - np.log10(amps_div)) > 0.2)
        sigma_div[bad] = 1e8

        # Iteratively fit and flag
        maxiter = 5
        niter = 0
        nflag = 0
        nflag_prev = -1
        while nflag != nflag_prev and niter < maxiter:
            p, bp_sp = _fit_bandpass(freqs, np.log10(amps_div), sigma_div, band)
            stdev_all = np.sqrt(np.average((bp_sp-np.log10(amps_div))**2, weights=(1/sigma_div)**2))
            stdev = min(maxStddev, stdev_all)
            bad = np.where(np.abs(bp_sp - np.log10(amps_div)) > nSigma*stdev)
            nflag = len(bad[0])
            if nflag == 0:
                break
            if niter > 0:
                nflag_prev = nflag
            sigma_div = sigma_orig.copy()  # reset flags to original ones
            sigma_div[bad] = 1e8
            niter += 1

        if plot:
            import matplotlib.pyplot as plt
            plt.plot(freqs, bp_sp, 'g-', lw=3)
            plt.plot(freqs, np.log10(amps_div), 'o', c='g')
            plt.plot(freqs[bad], np.log10(amps_div)[bad], 'o', c='r')
            plt.show()

        # Check whether entire station is bad (high stdev or high flagged fraction). If
        # so, flag all frequencies and polarizations
        if stdev_all > maxStddev * 5.0:
            # Station has high stddev relative to median bandpass
            print(('Flagged station {0} (pol {1}) due to high stddev '
                  '({2})'.format(s, pol, stdev_all)))
            weights[:, :, pol] = 0.0
        elif float(len(bad[0]))/float(len(freqs)) > maxFlaggedFraction:
            # Station has high fraction of flagged solutions
            print(('Flagged station {0} (pol {1}) due to high flagged fraction '
                  '({2})'.format(s, pol, float(len(bad[0]))/float(len(freqs)))))
            weights[:, :, pol] = 0.0
        elif median_val < median_min or median_val > median_max:
            # Station has extreme median value
            print(('Flagged station {0} (pol {1}) due to extreme median value '
                  '({2})'.format(s, pol, median_val)))
            weights[:, :, pol] = 0.0
        else:
            # Station is OK; flag solutions with high sigma values
            flagged = np.where(sigma_div > 1e3)
            nflagged_orig = len(np.where(weights[:, :, pol] == 0.0)[0])
            weights[:, flagged[0], pol] = 0.0
            nflagged_new = len(np.where(weights[:, :, pol] == 0.0)[0])
            prcnt = float(nflagged_new - nflagged_orig) / float(np.product(weights.shape[:-1])) * 100.0
            print(('Flagged {0}% of solutions for station {1} (pol {2})'.format(prcnt, s, pol)))

    outQueue.put([s, weights])


def run(soltab, chanWidth='', outSoltabName='bandpass', BadSBList = '', interpolate=True,
        removeTimeAxis=True, autoFlag=False, nSigma=5.0, maxFlaggedFraction=0.5,
        maxStddev=0.01, ncpu=0):
    """
    This operation for LoSoTo implements the Prefactor bandpass operation
    WEIGHT: flag-only compliant, no need for weight

    Parameters
    ----------
    chanWidth : str or float, optional
        the width of each channel in the data from which solutions were obtained. Can be
        either a string like "48kHz" or a float in Hz. If interpolate = True, chanWidth
        must be specified
    BadSBList : str, optional
        a list of bad subbands that will be flagged
    outSoltabName : str, optional
        Name of the output bandpass soltab. An existing soltab with this name will be
        overwritten
    interpolate : bool, optional
        If True, interpolate to a regular frequency grid and then smooth, ignoring bad
        subbands. If False, neither interpolation nor smoothing is done and the output
        frequency grid is the same as the input one. If interpolate = True, chanWidth
        must be specified
    removeTimeAxis : bool, optional
        If True, the time axis of the output bandpass soltab is removed by doing a median
        over time. If False, the output time grid is the same as the input one
    autoFlag : bool, optional
        If True, automatically flag bad frequencies and stations
    nSigma : float, optional
        Number of sigma for autoFlagging. Amplitudes outside of nSigma*stddev are flagged
    maxFlaggedFraction : float, optional
        Maximum allowable fraction of flagged frequencies for autoFlagging. Stations with
        higher fractions will be completely flagged
    maxStddev : float, optional
        Maximum allowable standard deviation for autoFlagging
    ncpu : int, optional
        Number of CPUs to use during autoFlagging (0 = all)
    """
    import numpy as np
    import scipy
    import scipy.ndimage

    logging.info("Running prefactor_bandpass on: "+soltab.name)
    solset = soltab.getSolset()

    solType = soltab.getType()
    if solType != 'amplitude':
       logging.warning("Soltab type of "+soltab.name+" is: "+solType+" should be amplitude. Ignoring.")
       return 1
    if soltab.name == outSoltabName and (removeTimeAxis or interpolate):
            logging.error("If removeTimeAxis = True or interpolate = True, outSoltabName must specify a new soltab.")
            raise ValueError("If removeTimeAxis = True or interpolate = True, outSoltabName must specify a new soltab.")

    if BadSBList == '':
      bad_sblist   = []
    else:
      bad_sblist = [int(SB) for SB in BadSBList.strip('\"\'').split(';')]


    for vals, weights, coord, selection in soltab.getValuesIter(returnAxes=['pol','ant','freq', 'time'], weight=True):
        vals = reorderAxes( vals, soltab.getAxesNames(), ['time', 'ant', 'freq', 'pol'] )
        weights = reorderAxes( weights, soltab.getAxesNames(), ['time', 'ant', 'freq', 'pol'] )
        pass

    amplitude_arraytmp = vals # axes are [time, ant, freq, pol]
    weights_arraytmp = weights # axes are [time, ant, freq, pol]
    flagged = np.where(amplitude_arraytmp == 1.0)
    weights_arraytmp[flagged] = 0.0
    nfreqs = len(soltab.freq[:])
    ntimes = len(soltab.time[:])
    nants = len(soltab.ant[:])

    subbandHz = 195.3125e3
    if interpolate:
        if chanWidth == '':
            logging.error("If interpolate = True, chanWidth must be specified.")
            raise ValueError("If interpolate = True, chanWidth must be specified.")
        if type(chanWidth) is str:
            letters = [1 for s in chanWidth[::-1] if s.isalpha()]
            indx = len(chanWidth) - sum(letters)
            unit = chanWidth[indx:]
            if unit.strip().lower() == 'hz':
                conversion = 1.0
            elif unit.strip().lower() == 'khz':
                conversion = 1e3
            elif unit.strip().lower() == 'mhz':
                conversion = 1e6
            else:
                logging.error("The unit on chanWidth was not understood.")
                raise ValueError("The unit on chanWidth was not understood.")
            chanWidthHz = float(chanWidth[:indx]) * conversion
        else:
            chanWidthHz = chanWidth
        offsetHz = subbandHz / 2.0 - 0.5 * chanWidthHz
        freqmin = np.min(soltab.freq[:]) + offsetHz # central frequency of first subband
        freqmax = np.max(soltab.freq[:]) + offsetHz # central frequency of last subband
        SBgrid = np.floor((soltab.freq[:]-np.min(soltab.freq[:]))/subbandHz)
        freqs_new  = np.arange(freqmin, freqmax+100e3, subbandHz)
        amps_array_flagged = np.zeros( (nants, ntimes, len(freqs_new), 2), dtype='float')
        amps_array = np.zeros( (nants, ntimes, len(freqs_new), 2), dtype='float')
        weights_array =  np.ones( (nants, ntimes, len(freqs_new), 2), dtype='float')

        logging.info("Have " + str(max(SBgrid)) + " subbands.")
        if len(freqs_new) < 20:
            logging.error("Frequency span is less than 20 subbands! The filtering will not work!")
            logging.error("Please run the calibrator pipeline on the full calibrator bandwidth.")
            raise ValueError("Frequency span is less than 20 subbands! Amplitude filtering will not work!")

        # make a mapping of new frequencies to old ones
        freq_mapping = {}
        for fn in freqs_new:
            ind = np.where(np.logical_and(soltab.freq < fn+subbandHz/2.0, soltab.freq >= fn-subbandHz/2.0))
            freq_mapping['{}'.format(fn)] = ind

    # remove bad subbands specified by user
    for bad_sb in bad_sblist:
        logging.info('Removing user-specified subband: ' + str(bad_sb))
        weights_arraytmp[:, :, bad_sb, :] = 0.0

    # remove bad solutions relative to the model bandpass
    if autoFlag:
        if ncpu == 0:
            import multiprocessing
            ncpu = multiprocessing.cpu_count()
        mpm = multiprocManager(ncpu, _flag_amplitudes)
        for s in range(nants):
            mpm.put([soltab.freq[:], amplitude_arraytmp[:, s, :, :], weights_arraytmp[:, s, :, :],
                     nSigma, maxFlaggedFraction, maxStddev, False, s])
        mpm.wait()
        for (s, w) in mpm.get():
            weights_arraytmp[:, s, :, :] = w

    # Now interpolate over flagged values and smooth over frequency and time axes
    if interpolate:
        for antenna_id in range(len(soltab.ant[:])):
            for time in range(len(soltab.time[:])):
                amp_xx_tmp = np.copy(amplitude_arraytmp[time, antenna_id, :, 0])
                amp_yy_tmp = np.copy(amplitude_arraytmp[time, antenna_id, :, 1])
                freq_tmp = soltab.freq[:]
                assert len(amp_xx_tmp[:]) == len(freq_tmp[:])
                mask_xx = np.not_equal(weights_arraytmp[time, antenna_id, :, 0], 0.0)
                if np.sum(mask_xx)>2:
                    amps_xx_tointer = amp_xx_tmp[mask_xx]
                    freq_xx_tointer = freq_tmp[mask_xx]
                    amps_array_flagged[antenna_id, time, :, 0] = np.interp(freqs_new, freq_xx_tointer, amps_xx_tointer)
                elif time>0:
                    amps_array_flagged[antenna_id, time, :, 0] = amps_array_flagged[antenna_id, (time-1), :, 0]
                mask_yy = np.not_equal(weights_arraytmp[time, antenna_id, :, 1], 0.0)
                if np.sum(mask_yy)>2:
                    amps_yy_tointer = amp_yy_tmp[mask_yy]
                    freq_yy_tointer = freq_tmp[mask_yy]
                    amps_array_flagged[antenna_id, time, :, 1] = np.interp(freqs_new, freq_yy_tointer, amps_yy_tointer)
                elif time>0:
                    amps_array_flagged[antenna_id, time, :, 1] = amps_array_flagged[antenna_id, (time-1), :, 1]

        ampsoutfile = open('calibrator_amplitude_array.txt','w')
        ampsoutfile.write('# Antenna name, Antenna ID, subband, XXamp, YYamp, frequency\n')
        for antenna_id in range(len(soltab.ant[:])):
            if np.all(weights_arraytmp[:, antenna_id, :, :] == 0.0):
                weights_array[antenna_id, :, :, :] = 0.0
            else:
                amp_xx = np.copy(amps_array_flagged[antenna_id, :, :, 0])
                amp_yy = np.copy(amps_array_flagged[antenna_id, :, :, 1])

                amp_xx = scipy.ndimage.filters.median_filter(amp_xx, (3,3))
                amp_xx = scipy.ndimage.filters.median_filter(amp_xx, (7,1))
                amp_yy = scipy.ndimage.filters.median_filter(amp_yy, (3,3))
                amp_yy = scipy.ndimage.filters.median_filter(amp_yy, (7,1))

                for i in range(len(freqs_new)):
                    ampsoutfile.write('%s %s %s %s %s %s\n'%(soltab.ant[antenna_id], antenna_id,
                                                             i, np.median(amp_xx[:,i], axis=0),
                                                             np.median(amp_yy[:,i], axis=0),
                                                             freqs_new[i]))

                for time in range(len(soltab.time[:])):
                    amps_array[antenna_id, time, :, 0] = np.copy(_savitzky_golay(amp_xx[time,:], 17, 2))
                    amps_array[antenna_id, time, :, 1] = np.copy(_savitzky_golay(amp_yy[time,:], 17, 2))

                for i in range(len(freqs_new)):
                    amps_array[antenna_id, :, i, 0] = np.median(amps_array[antenna_id, :, i, 0])
                    amps_array[antenna_id, :, i, 1] = np.median(amps_array[antenna_id, :, i, 1])
                    ind = freq_mapping['{}'.format(freqs_new[i])]
                    for p in range(2):
                        # If half or more of original frequencies are flagged, flag the
                        # output frequency as well
                        nflagged = len(np.where(weights_arraytmp[:, antenna_id, ind, p] == 0.0)[0])
                        ntot = weights_arraytmp.shape[0] * len(ind[0])
                        if ntot > 0:
                            if float(nflagged)/float(ntot) >= 0.5:
                                weights_array[antenna_id, :, i, p] = 0.0
        amps_array = amps_array.swapaxes(0, 1)
        weights_array = weights_array.swapaxes(0, 1)
    else:
        amps_array = amplitude_arraytmp
        weights_array = weights_arraytmp
        freqs_new = soltab.freq[:]

    # delete existing bandpass soltab if needed and write solutions
    if soltab.name != outSoltabName:
        try:
            new_soltab = solset.getSoltab(outSoltabName)
            new_soltab.delete()
        except:
            pass

    if removeTimeAxis:
        # Write bandpass, taking median over the time axis
        new_soltab = solset.makeSoltab(soltype='amplitude', soltabName=outSoltabName,
                                 axesNames=['ant', 'freq', 'pol'],
                                 axesVals=[soltab.ant, freqs_new, ['XX', 'YY']],
                                 vals=np.median(amps_array, axis=0),
                                 weights=np.median(weights_array, axis=0))
        new_soltab.addHistory('CREATE (by PREFACTOR_BANDPASS operation) with BadSBList = {0}, '
                                  'interpolate={1}, removeTimeAxis={2}, autoFlag={3}, nSigma={4}, '
                                  'maxFlaggedFraction={5}, maxStddev={6}'.format(BadSBList,
                                  interpolate, removeTimeAxis, autoFlag, nSigma,
                                  maxFlaggedFraction, maxStddev))
    else:
        # Write bandpass, preserving the time axis
        if soltab.name == outSoltabName:
            soltab.setValues(amps_array)
            soltab.setValues(weights_array, weight=True)
            soltab.addHistory('BANDPASS processed with BadSBList = {0}, interpolate={1}, '
                              'removeTimeAxis={2}, autoFlag={3}, nSigma={4}, '
                              'maxFlaggedFraction={5}, maxStddev={6}'.format(BadSBList,
                              interpolate, removeTimeAxis, autoFlag, nSigma,
                              maxFlaggedFraction, maxStddev))
        else:
            new_soltab = solset.makeSoltab(soltype='amplitude', soltabName=outSoltabName,
                                     axesNames=['time', 'ant', 'freq', 'pol'],
                                     axesVals=[soltab.time, soltab.ant, freqs_new, ['XX', 'YY']],
                                     vals=amps_array, weights=weights_array)
            new_soltab.addHistory('CREATE (by PREFACTOR_BANDPASS operation) with BadSBList = {0}, '
                                  'interpolate={1}, removeTimeAxis={2}, autoFlag={3}, nSigma={4}, '
                                  'maxFlaggedFraction={5}, maxStddev={6}'.format(BadSBList,
                                  interpolate, removeTimeAxis, autoFlag, nSigma,
                                  maxFlaggedFraction, maxStddev))

    return 0



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from losoto.lib_operations import *
from losoto._logging import logger as logging

logging.debug('Loading SMOOTH module.')

def _run_parser(soltab, parser, step):
    axesToSmooth = parser.getarraystr( step, 'axesToSmooth' ) # no default
    size = parser.getarrayint( step, 'size', [] )
    mode = parser.getstr( step, 'mode', 'runningmedian' )
    degree = parser.getint( step, 'degree', 1 )
    replace = parser.getbool( step, 'replace', False )
    log = parser.getbool( step, 'log', False )
    refAnt = parser.getstr( step, 'refAnt', '' )

    parser.checkSpelling( step, soltab, ['axesToSmooth', 'size', 'mode', 'degree', 'replace', 'log', 'refAnt'])
    return run(soltab, axesToSmooth, size, mode, degree, replace, log, refAnt)


def _savitzky_golay(y, window_size, order):
    from scipy.signal import savgol_filter

    # replace any NaNs using linear interpolation
    nans = np.isnan(y)
    if np.any(nans):
        x = np.array(range(len(y)))
        y_nonan = np.interp(x, x[~nans], y[~nans])
    else:
        y_nonan = y

    y_filt = savgol_filter(y_nonan, window_size, order)

    # put any NaNs back
    if np.any(nans):
        y_filt[nans] = np.nan

    return y_filt


def run( soltab, axesToSmooth, size=[], mode='runningmedian', degree=1, replace=False, log=False, refAnt=''):
    """
    A smoothing function: running-median on an arbitrary number of axes, running polyfit and Savitzky-Golay on one axis, or set all solutions to the mean/median value.
    WEIGHT: flag ready.

    Parameters
    ----------
    axesToSmooth : array of str
        Axes used to compute the smoothing function.

    size : array of int, optional
        Window size for the runningmedian, savitzky-golay, and runningpoly (array of same size of axesToSmooth), by default [].

    mode : {'runningmedian','runningpoly','savitzky-golay','mean','median'}, optional
        Runningmedian or runningpoly or Savitzky-Golay or mean or median (these last two values set all the solutions to the mean/median), by default "runningmedian".

    degree : int, optional
        Degrees of the polynomia for the runningpoly or savitzky-golay modes, by default 1.

    replace : bool, optional
        Flagged data are replaced with smoothed value and unflagged, by default False.

    log : bool, optional
        clip is done in log10 space, by default False

    refAnt : str, optional
        Reference antenna for phases. By default None.
    """

    import numpy as np
    from scipy.ndimage import generic_filter

    if refAnt == '': refAnt = None
    elif not refAnt in soltab.getAxisValues('ant', ignoreSelection = True):
        logging.warning('Reference antenna '+refAnt+' not found. Using: '+soltab.getAxisValues('ant')[1])
        refAnt = soltab.getAxisValues('ant')[1]

    if mode == "runningmedian" and len(axesToSmooth) != len(size):
        logging.error("Axes and Size lengths must be equal for runningmedian.")
        return 1

    if (mode == "runningpoly" or mode=="savitzky-golay") and (len(axesToSmooth) != 1 or len(size) != 1):
        logging.error("Axes and size lengths must be 1 for runningpoly or savitzky-golay.")
        return 1

    if (mode == "runningpoly" or mode=="savitzky-golay") and soltab.getType() == 'phase':
        logging.error("Runningpoly and savitzky-golay modes cannot work on phases.")
        return 1

    for i, s in enumerate(size):
        if s % 2 == 0:
            logging.warning('Size should be odd, adding 1.')
            size[i] += 1

    logging.info("Smoothing soltab: "+soltab.name)

    for i, axis in enumerate(axesToSmooth[:]):
        if axis not in soltab.getAxesNames():
            del axesToSmooth[i]
            del size[i]
            logging.warning('Axis \"'+axis+'\" not found. Ignoring.')

    if soltab.getType() == 'amplitude' and not log:
        logging.warning('Amplitude solution tab detected and log=False. Amplitude solution tables should be treated in log space.')

    if mode == 'median' or mode == 'mean':
        vals = soltab.getValues(retAxesVals=False, refAnt=refAnt)
        if log: vals = np.log10(vals)
        weights = soltab.getValues(retAxesVals=False, weight=True)
        np.putmask(vals, weights==0, np.nan)
        idx_axes = [soltab.getAxesNames().index(axisToSmooth) for axisToSmooth in axesToSmooth]

        # handle phases by using a complex array
        if soltab.getType() == 'phase':
            vals = np.exp(1j*vals)

        if mode == 'median':
            vals[:] = np.nanmedian( vals, axis=idx_axes, keepdims=True)
        if mode == 'mean':
            logging.warning('Mean does not support NaN yet, use median if it is a problem.')
            vals[:] = np.mean( vals, axis=idx_axes, keepdims=True) # annoying np.nanmean does not accept axis=list!

        # go back to phases
        if soltab.getType() == 'phase':
            vals = np.angle(vals)

        # write back
        if log: vals = 10**vals
        soltab.setValues(vals)
        if replace:
            weights[ (weights == 0) ] = 1
            weights[ np.isnan(vals) ] = 0 # all the slice was flagged, cannot estrapolate value
            soltab.setValues(weights, weight=True)

    else:
        for vals, weights, coord, selection in soltab.getValuesIter(returnAxes=axesToSmooth, weight=True, refAnt=refAnt):

            # skip completely flagged selections
            if (weights == 0).all(): continue
            if log: vals = np.log10(vals)

            if mode == 'runningmedian':
                vals_bkp = vals[ weights == 0 ]

                # handle phases by using a complex array
                if soltab.getType() == 'phase':
                    vals = np.exp(1j*vals)

                    valsreal = np.real(vals)
                    valsimag = np.imag(vals)
                    np.putmask(valsreal, weights == 0, np.nan)
                    np.putmask(valsimag, weights == 0, np.nan)

                    # run generic_filter twice, once for real once for imaginary
                    valsrealnew = generic_filter(valsreal, np.nanmedian, size=size, mode='constant', cval=np.nan)
                    valsimagnew = generic_filter(valsimag, np.nanmedian, size=size, mode='constant', cval=np.nan)
                    valsnew = valsrealnew + 1j*valsimagnew # go back to complex
                    valsnew = np.angle(valsnew) # go back to phases

                else: # other than phases
                    np.putmask(vals, weights == 0, np.nan)
                    valsnew = generic_filter(vals, np.nanmedian, size=size, mode='constant', cval=np.nan)


                if replace:
                    weights[ weights == 0] = 1
                    weights[ np.isnan(valsnew) ] = 0 # all the size was flagged cannoth estrapolate value
                else:
                    valsnew[ weights == 0 ] = vals_bkp

            elif mode == 'runningpoly':
                def polyfit(data):
                    if (np.isnan(data)).all(): return np.nan # all size is flagged
                    x = np.arange(len(data))[ ~np.isnan(data)]
                    y = data[ ~np.isnan(data) ]
                    p = np.polynomial.polynomial.polyfit(x, y, deg=degree)
                    #import matplotlib as mpl
                    #mpl.use("Agg")
                    #import matplotlib.pyplot as plt
                    #plt.plot(x, y, 'ro')
                    #plt.plot(x, np.polyval( p[::-1], x ), 'k-')
                    #plt.savefig('test.png')
                    #sys.exit()
                    return np.polyval( p[::-1], (size[0]-1)/2 ) # polyval has opposite convention for polynomial order

                # flags and at edges pass 0 and then remove them
                vals_bkp = vals[ weights == 0 ]
                np.putmask(vals, weights==0, np.nan)
                valsnew = generic_filter(vals, polyfit, size=size[0], mode='constant', cval=np.nan)
                if replace:
                    weights[ weights == 0] = 1
                    weights[ np.isnan(valsnew) ] = 0 # all the size was flagged cannot extrapolate value
                else:
                    valsnew[ weights == 0 ] = vals_bkp
                #print coord['ant'], vals, valsnew

            elif mode == 'savitzky-golay':
                vals_bkp = vals[ weights == 0 ]
                np.putmask(vals, weights==0, np.nan)
                valsnew = _savitzky_golay(vals, size[0], degree)
                if replace:
                    weights[ weights == 0] = 1
                    weights[ np.isnan(valsnew) ] = 0 # all the size was flagged cannot extrapolate value
                else:
                    valsnew[ weights == 0 ] = vals_bkp

            else:
                logging.error('Mode must be: runningmedian, runningpoly, savitzky-golay, median or mean')
                return 1

            if log: valsnew = 10**valsnew
            soltab.setValues(valsnew, selection)
            if replace: soltab.setValues(weights, selection, weight=True)

    soltab.flush()
    soltab.addHistory('SMOOTH (over %s with mode = %s)' % (axesToSmooth, mode))
    return 0



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from losoto.lib_operations import *
from losoto._logging import logger as logging

logging.debug('Loading REFERENCE module.')

def _run_parser(soltab, parser, step):
    refAnt = parser.getstr( step, 'refAnt', '' )
    refDir = parser.getstr( step, 'refDir', '' )

    parser.checkSpelling( step, soltab, ['refAnt','refDir'])
    return run(soltab, refAnt, refDir)


def run( soltab, refAnt='', refDir=''):
    """
    Reference to an antenna

    Parameters
    ----------
    refAnt : str, optional
        Reference antenna for phases or "best" to use the least flagged antenna. Empty string does not change phases. Default: ''.
    refDir : str, optional
        Reference direction for phases. Empty string does not change phases. Default: ''.
    """

    if not soltab.getType() in ['phase', 'scalarphase', 'rotation', 'tec', 'clock', 'tec3rd', 'rotationmeasure']:
        logging.error('Reference possible only for phase, scalarphase, clock, tec, tec3rd, rotation and rotationmeasure solution tables. Ignore referencing.')
        return 1

    if refAnt == 'best': 
        weights = soltab.getValues(retAxesVals=False, weight=True)
        weights = np.sum(weights, axis=tuple([i for i, axis_name in enumerate(soltab.getAxesNames()) if axis_name != 'ant']), dtype=np.float)
        refAnt = soltab.getAxisValues('ant')[np.where(weights == np.max(weights))[0][0]]
        logging.info('Using %s for reference antenna.' % refAnt)

    elif not refAnt in soltab.getAxisValues('ant', ignoreSelection = True) and refAnt != '':
        logging.error('Reference antenna '+refAnt+' not found.')
        return 1

    if not refDir in soltab.getAxisValues('dir', ignoreSelection = True) and refDir != '':
        logging.error('Reference direction '+refDir+' not found.')
        return 1

    # get reference direction if needed
    if refDir != '' and refAnt != '':
        logging.info('Referencing on both dir and ant.')
        soltab.setSelection(dir=refDir, ant=refAnt)
        valsRef = soltab.getValues(retAxesVals=False)
        soltab.clearSelection()
        vals = soltab.getValues(retAxesVals=False)
        dirAxis = soltab.getAxesNames().index('dir')
        antAxis = soltab.getAxesNames().index('ant')
        print ('len vals ref', valsRef.shape)
        valsRef = np.repeat(valsRef, axis=dirAxis, repeats=len(soltab.getAxisValues('dir')))
        valsRef = np.repeat(valsRef, axis=antAxis, repeats=len(soltab.getAxisValues('ant')))
        vals = vals - valsRef

    elif refDir != '':
        soltab.setSelection(dir=refDir)
        valsRefDir = soltab.getValues(retAxesVals=False)
        soltab.clearSelection()
        vals = soltab.getValues(retAxesVals=False)
        dirAxis = soltab.getAxesNames().index('dir')
        vals = vals - np.repeat(valsRefDir, axis=dirAxis, repeats=len(soltab.getAxisValues('dir')))

    # use automatic antenna referencing
    elif refAnt != '':
        vals = soltab.getValues(retAxesVals=False, refAnt=refAnt)

    soltab.setValues(vals)

    soltab.addHistory('REFERENCED (to antenna: %s)' % (refAnt))
    return 0



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
This tool is used to convert a parmdb into a H5parm format.
It can be run on a globaldb created with parmdb_collector.py
or on a single SB which contains the necessary
sky/instrument/ANTENNA/FIELD tables.
It can also load a dictionary from a json or pickle file with the keys 
instrumentdb (array with file names), antenna, field and skydb containing the 
corresponding file names.
It handles Gain/DirectionalGain/RotationAngle/
           Clock/TEC/CommonRotationAngle/CommonScalarPhase/CommonScalarAmpitude solution types.
"""

import sys, os
import socket
import numpy as np
import pyrap.tables as pt
from losoto import _version
from losoto.h5parm import h5parm as h5parm_mod
try:
    from . import progressbar
except ImportError:
    import losoto.progressbar as progressbar
from losoto._logging import logger as logging

def parmdbToAxes(solEntry):
    """
    Extract the information written as a string in the parmdb format
    """
    pol = None
    pol1 = None
    pol2 = None
    dir = None
    ant = None
    parm = None

    thisSolType = solEntry.split(':')[0]

    # For CommonRotationAngle assuming [CommonRotationAngle:ant]
    if thisSolType == 'CommonRotationAngle':
        thisSolType, ant = solEntry.split(':')
        dir = 'pointing'

    # For RotationAngle assuming [RotationAngle:ant:sou]
    elif thisSolType == 'RotationAngle':
        thisSolType, ant, dir = solEntry.split(':')

    # For RotationMeasure assuming [RotationMeasure:ant:sou]
    elif thisSolType == 'RotationMeasure':
        dir = 'pointing'
        try:
            thisSolType, ant = solEntry.split(':')
        except:
            thisSolType, ant, dir = solEntry.split(':')

    # For TEC assuming [TEC:ant or TEC:pol:ant]
    elif thisSolType == 'TEC':
        try:
            thisSolType, ant = solEntry.split(':')
        except:
            thisSolType, pol, ant = solEntry.split(':')
            pol1 = pol
            pol2 = pol
        dir = 'pointing'

    # For Clock assuming [Clock:ant or Clock:pol:ant]
    elif thisSolType == 'Clock':
        try:
            thisSolType, ant = solEntry.split(':')
        except:
            thisSolType, pol, ant = solEntry.split(':')
            pol1 = pol
            pol2 = pol

    # For CommonScalarPhase assuming [CommonScalarPhase:ant]
    elif thisSolType == 'CommonScalarPhase':
        thisSolType, ant = solEntry.split(':')
        dir = 'pointing'

    # For CommonScalarPhase assuming [CommonScalarPhase:ant]
    elif thisSolType == 'CommonScalarAmplitude':
        thisSolType, ant = solEntry.split(':')
        dir = 'pointing'

    # For ScalarPhase assuming [ScalarPhase:ant:sou]
    elif thisSolType == 'ScalarPhase':
        thisSolType, ant, dir = solEntry.split(':')

    # For ScalarPhase assuming [ScalarAmplitude:ant:sou]
    elif thisSolType == 'ScalarAmplitude':
        thisSolType, ant, dir = solEntry.split(':')

    # For Gain assuming [Gain:pol1:pol2:parm:ant]
    elif thisSolType == 'Gain':
        thisSolType, pol1, pol2, parm, ant = solEntry.split(':')
        dir = 'pointing'

    # For DirectionalGain assuming [DirecitonalGain:pol1:pol2:parm:ant:sou]
    elif thisSolType == 'DirectionalGain':
        thisSolType, pol1, pol2, parm, ant, dir = solEntry.split(':')

    else:
        logging.error('Unknown solution type "'+thisSolType+'". Ignored.')

    if pol1 is not None and pol2 is not None:
        if pol1 == '0' and pol2 == '0': pol = 'XX'
        if pol1 == '1' and pol2 == '0': pol = 'YX'
        if pol1 == '0' and pol2 == '1': pol = 'XY'
        if pol1 == '1' and pol2 == '1': pol = 'YY'

    return pol, dir, ant, parm

def getValuesGrid(parmdb,solType):
    '''Get values grid mimics the behaviour of getValues grid of lofar.parmdb. 
    We assume the data is stored as a regular grid in frequency and time. '''
    names = pt.table(parmdb+"/NAMES").getcol("NAME")
    indices = [i for i,j in enumerate(names) if re.match(solType,name)]
    data = {}
    for idx in indices:
         newt = pt.taql("SELECT * from $parmdb where NAMEID == $idx")
         data[names[idx]] = {}
         freqsc = [newt.getcol("STARTX")[0], newt.getcol("ENDX")[0]]
         timesc = [newt.getcol("STARTY"), newt.getcol("ENDY")]
         values = newt.getcol("VALUES")
         ntimes = values.shape[0]*values.shape[1]
         nfreq = values.shape[2]
         dfreq = (freqsc[1]-freqsc[0])/nfreq
         freqs = np.linspace(freqsc[0]+0.5*dfreq,freqsc[1]+0.5*dfreq, nfreq, 
                             endpoint=False)
         freqwidths = dfreq*np.ones_like(freqs)
         dtimes = (timesc[-1][-1] - timesc[0][0])/ntimes
         times = np.linspace(timesc[0][0]+0.5*dtimes,timesc[-1][-1]+0.5*dtimes, 
                             ntimes, endpoint=False)
         timewidths = dtimes*np.ones_like(times)
         data[names[idx]]['freqs'] = freqs
         data[names[idx]]['freqwidths'] = freqwidths
         data[names[idx]]['times'] = times
         data[names[idx]]['timewidths'] = timewidths
         data[names[idx]]['values'] = values
    return data
         
         
def create_h5parm(instrumentdbFiles, antennaFile, fieldFile, skydbFile,
                  h5parmFile, complevel, solsetName, globaldbFile=None, verbose=False):
    """
    Create the h5parm file.
    Input:
       instrumentdbFiles - list of the finenames of the solutions.
       antennaFile - file name of the antenna table.
       fieldFile - file name of the field table.
       skydbFile - file name of the sky table.
       h5parmFile - file name of the h5parm file that will be created.
       complevel - level of compression. It is usually 5.
       solsetName - Name of the solution set. Usually "sol###".
       globaldbFile (optional) - Name of the globaldbFile. Used only for 
         logging purposes.
    """
    
    # open/create the h5parm file and the solution-set
    h5parm = h5parm_mod(h5parmFile, readonly = False, complevel = complevel)
    
    solset = h5parm.makeSolset(solsetName)
    
    # Create tables using the first instrumentdb
    # TODO: all the instrument tables should be checked
    #pdb = lofar.parmdb.parmdb(instrumentdbFiles[0])
    #    solTypes = list(set(x[0] for x in  (x.split(":") for x in pdb.getNames())))
    names = pt.table(instrumentdbFiles[0]+"/NAMES").getcol("NAME")
    
    solTypes = list(set(x[0] for x in  (x.split(":") for x in names)))
    logging.info('Found solution types: '+', '.join(solTypes))

    # rewrite solTypes in order to put together
    # Gain <-> DirectionalGain
    # CommonRotationAngle <-> RotationAngle
    # CommonScalarPhase <-> ScalarPhase
    # CommonScalarAmplitude <-> ScalarAmplitude
    # it also separate Real/Imag/Ampl/Phase into different solTypes
    if "Gain" in solTypes:
        solTypes.remove('Gain')
        solTypes.append('.*Gain:.*:Real')
        solTypes.append('.*Gain:.*:Imag')
        solTypes.append('.*Gain:.*:Ampl')
        solTypes.append('.*Gain:.*:Phase')
    if "DirectionalGain" in solTypes:
        solTypes.remove('DirectionalGain')
        solTypes.append('.*Gain:.*:Real')
        solTypes.append('.*Gain:.*:Imag')
        solTypes.append('.*Gain:.*:Ampl')
        solTypes.append('.*Gain:.*:Phase')
    if "RotationAngle" in solTypes:
        solTypes.remove('RotationAngle')
        solTypes.append('.*RotationAngle')
    if "CommonRotationAngle" in solTypes:
        solTypes.remove('CommonRotationAngle')
        solTypes.append('.*RotationAngle')
    if "RotationMeasure" in solTypes:
        solTypes.remove('RotationMeasure')
        solTypes.append('.*RotationMeasure')
    if "ScalarPhase" in solTypes:
        solTypes.remove('ScalarPhase')
        solTypes.append('.*ScalarPhase')
    if "CommonScalarPhase" in solTypes:
        solTypes.remove('CommonScalarPhase')
        solTypes.append('.*ScalarPhase')
    if "CommonScalarAmplitude" in solTypes:
        solTypes.remove('CommonScalarAmplitude')
        solTypes.append('.*ScalarAmplitude')
    solTypes = list(set(solTypes))

    # every soltype creates a different solution-table
    for solType in solTypes:

        # skip missing solTypes (not all parmdbs have e.g. TEC)
        #if len(pdb.getNames(solType+':*')) == 0: continue
        
        found = False
        for name in names:
            if re.match(solType,name):
                found = True
                break
        if not found:
            continue
                        
        pols = set() 
        dirs = set() 
        ants = set()
        freqs = set() 
        times = set() 
        ptype = set()

        logging.info('Reading '+solType+'.')

        pbar = progressbar.ProgressBar(maxval=len(instrumentdbFiles)).start()
        ipbar = 0

        for instrumentdbFile in sorted(instrumentdbFiles):

            #pdb = lofar.parmdb.parmdb(instrumentdbFile)
            
            # create the axes grid, necessary if not all entries have the same axes lenght
            #data = pdb.getValuesGrid(solType+':*')
            data = getValuesGrid(instrumentdbFile,solType + ".*")
            # check good instrument table
            if len(data) == 0:
                logging.error('Instrument table %s is empty, ignoring.' % instrumentdbFile)

            for solEntry in data:

                pol, dir, ant, parm = parmdbToAxes(solEntry)
                if pol is not None: pols |= set([pol])
                if dir is not None: dirs |= set([dir])
                if ant is not None: ants |= set([ant])
                freqs |= set(data[solEntry]['freqs'])
                times |= set(data[solEntry]['times'])
                pbar.update(ipbar)
            ipbar += 1

        pbar.finish()

        pols = np.sort(list(pols)) 
        dirs = np.sort(list(dirs)) 
        ants = np.sort(list(ants)) 
        freqs = np.sort(list(freqs)) 
        times = np.sort(list(times))
        shape = [i for i in (len(pols), len(dirs), len(ants), len(freqs), len(times)) if i != 0]
        vals = np.empty(shape)
        vals[:] = np.nan
        weights = np.zeros(shape, dtype=np.float16)

        logging.info('Filling table.')
        pbar = progressbar.ProgressBar(maxval=len(instrumentdbFiles)).start()
        ipbar = 0

        for instrumentdbFile in instrumentdbFiles:

            #pdb = lofar.parmdb.parmdb(instrumentdbFile)

            # fill the values
            #data = pdb.getValuesGrid(solType+':*')
            data = getValuesGrid(instrumentdbFile,solType + ".*")
            #if 'Real' in solType: dataIm = pdb.getValuesGrid(solType.replace('Real','Imag')+':*')
            #if 'Imag' in solType: dataRe = pdb.getValuesGrid(solType.replace('Imag','Real')+':*')
            if 'Real' in solType: dataIm = getValuesGrid(instrumentdbFile,
                                                         solType.replace('Real','Imag')+':.*')
            if 'Imag' in solType: dataRe = getValuesGrid(instrumentdbFile,
                                                         solType.replace('Imag','Real')+':.*')
            for solEntry in data:

                pol, dir, ant, parm = parmdbToAxes(solEntry)
                ptype |= set([solEntry.split(':')[0]]) # original parmdb solution type

                freq = data[solEntry]['freqs']
                time = data[solEntry]['times']

                val = data[solEntry]['values']

                # convert Real and Imag in Amp and Phase respectively
                if parm == 'Real':
                    solEntryIm = solEntry.replace('Real','Imag')
                    valI = dataIm[solEntryIm]['values']
                    val = np.sqrt((val**2)+(valI**2))
                if parm == 'Imag':
                    solEntryRe = solEntry.replace('Imag','Real')
                    valR = dataRe[solEntryRe]['values']
                    val = np.arctan2(val, valR)

                coords = []
                if pol is not None:
                    polCoord = np.searchsorted(pols, pol)
                    coords.append(polCoord)
                if dir is not None:
                    dirCoord = np.searchsorted(dirs, dir)
                    coords.append(dirCoord)
                if ant is not None:
                    antCoord = np.searchsorted(ants, ant)
                    coords.append(antCoord)
                freqCoord = np.searchsorted(freqs, freq)
                timeCoord = np.searchsorted(times, time)
                vals[tuple(coords)][np.ix_(freqCoord,timeCoord)] = val.T
                weights[tuple(coords)][np.ix_(freqCoord,timeCoord)] = 1
                pbar.update(ipbar)
            ipbar += 1

        np.putmask(vals, ~np.isfinite(vals), 0) # put inf and nans to 0
        #vals = np.nan_to_num(vals) # replace nans with 0 (flagged later)

        pbar.finish()
        if solType == '*RotationAngle':
            np.putmask(weights, vals == 0., 0) # flag where val=0
            solset.makeSoltab('rotation', axesNames=['dir','ant','freq','time'], \
                    axesVals=[dirs,ants,freqs,times], vals=vals, weights=weights, parmdbType=', '.join(list(ptype)))
        if solType == '*RotationMeasure':
            np.putmask(weights, vals == 0., 0) # flag where val=0
            solset.makeSoltab('rotationmeasure', axesNames=['dir','ant','freq','time'], \
                    axesVals=[dirs,ants,freqs,times], vals=vals, weights=weights, parmdbType=', '.join(list(ptype)))
        elif solType == '*ScalarPhase':
            np.putmask(weights, vals == 0., 0)
            solset.makeSoltab('scalarphase', axesNames=['dir','ant','freq','time'], \
                    axesVals=[dirs,ants,freqs,times], vals=vals, weights=weights, parmdbType=', '.join(list(ptype)))
        elif solType == '*ScalarAmplitude':
            np.putmask(weights, vals == 0., 0)
            solset.makeSoltab('scalaramplitude', axesNames=['dir','ant','freq','time'], \
                    axesVals=[dirs,ants,freqs,times], vals=vals, weights=weights, parmdbType=', '.join(list(ptype)))
        elif solType == 'Clock':
            np.putmask(weights, vals == 0., 0)
            # clock may be diag or scalar
            if len(pols) == 0:
                solset.makeSoltab('clock', axesNames=['ant','freq','time'], \
                    axesVals=[ants,freqs,times], vals=vals, weights=weights, parmdbType=', '.join(list(ptype)))
            else:
                solset.makeSoltab('clock', axesNames=['pol','ant','freq','time'], \
                    axesVals=[pol,ants,freqs,times], vals=vals, weights=weights, parmdbType=', '.join(list(ptype)))
        elif solType == 'TEC':
            np.putmask(weights, vals == 0., 0)
            # tec may be diag or scalar
            if len(pols) == 0:
                solset.makeSoltab('tec', axesNames=['dir','ant','freq','time'], \
                    axesVals=[dirs,ants,freqs,times], vals=vals, weights=weights, parmdbType=', '.join(list(ptype)))
            else:
                solset.makeSoltab('tec', axesNames=['pol','dir','ant','freq','time'], \
                    axesVals=[pols,dirs,ants,freqs,times], vals=vals, weights=weights, parmdbType=', '.join(list(ptype)))
        elif solType == '*Gain:*:Real' or solType == '*Gain:*:Ampl':
            np.putmask(vals, vals == 0, 1) # nans were put to 0 before, set them to 1
            np.putmask(weights, vals == 1., 0) # flag where val=1
            solset.makeSoltab('amplitude', axesNames=['pol','dir','ant','freq','time'], \
                    axesVals=[pols,dirs,ants,freqs,times], vals=vals, weights=weights, parmdbType=', '.join(list(ptype)))
        elif solType == '*Gain:*:Imag' or solType == '*Gain:*:Phase':
            np.putmask(weights, vals == 0., 0) # falg where val=0
            solset.makeSoltab('phase', axesNames=['pol','dir','ant','freq','time'], \
                    axesVals=[pols,dirs,ants,freqs,times], vals=vals, weights=weights, parmdbType=', '.join(list(ptype)))

        logging.info('Flagged data: %.3f%%' % (100.*(len(weights.flat)-np.count_nonzero(weights))/len(weights.flat)))

    logging.info('Collecting information from the ANTENNA table.')
    antennaTable = pt.table(antennaFile, ack=False)
    antennaNames = antennaTable.getcol('NAME')
    antennaPositions = antennaTable.getcol('POSITION')
    antennaTable.close()
    antennaTable = solset.obj._f_get_child('antenna')
    antennaTable.append(list(zip(*(antennaNames,antennaPositions))))

    logging.info('Collecting information from the FIELD table.')
    fieldTable = pt.table(fieldFile, ack=False)
    phaseDir = fieldTable.getcol('PHASE_DIR')
    pointing = phaseDir[0, 0, :]
    fieldTable.close()

    sourceTable = solset.obj._f_get_child('source')
    # add the field centre, that is also the direction for Gain and Common*
    sourceTable.append([('pointing',pointing)])

    dirs = []
    for tab in solset.obj._v_children:
        c = solset.obj._f_get_child(tab)
        if c._v_name != 'antenna' and c._v_name != 'source':
            if c.__contains__('dir'):
                dirs.extend(list(set(c.dir)))
    # remove duplicates
    dirs = list(set(dirs))
    # remove any pointing (already in the table)
    if 'pointing' in dirs:
        dirs.remove('pointing')

    if not os.path.isdir(skydbFile) and dirs!=[]:
        logging.critical('Missing skydb table.')
        sys.exit(1)

    if dirs != []:
        logging.info('Collecting information from the sky table.')
        sourceFile = skydbFile + '/SOURCES'
        src_table = pt.table(sourceFile, ack=False)
        sub_tables = src_table.getsubtables()
        vals = []
        ra = dec = np.nan
        has_patches_subtable = False
        for sub_table in sub_tables:
            if 'PATCHES' in sub_table:
                has_patches_subtable = True
        if has_patches_subtable:
            # Read values from PATCHES subtable
            src_table.close()
            sourceFile = skydbFile + '/SOURCES/PATCHES'
            src_table = pt.table(sourceFile, ack=False)
            patch_names = src_table.getcol('PATCHNAME')
            patch_ras = src_table.getcol('RA')
            patch_decs = src_table.getcol('DEC')
            for source in dirs:
                try:
                    patch_indx = patch_names.index(source)
                    ra = patch_ras[patch_indx]
                    dec = patch_decs[patch_indx]
                except ValueError:
                    ra = np.nan
                    dec = np.nan
                    logging.error('Cannot find the source '+source+'. I leave NaNs.')
                vals.append([ra, dec])
            src_table.close()
        else:
            # Try to read default values from parmdb instead
            skydb = lofar.parmdb.parmdb(skydbFile)
            vals = []
            ra = dec = np.nan

            for source in dirs:
                try:
                    ra = skydb.getDefValues('Ra:' + source)['Ra:' + source][0][0]
                    dec = skydb.getDefValues('Dec:' + source)['Dec:' + source][0][0]
                except KeyError:
                    # Source not found in skymodel parmdb, try to find components
                    logging.warning('Cannot find the source '+source+'. Trying components.')
                    ra = np.array(list(skydb.getDefValues('Ra:*' + source + '*').values()))
                    dec = np.array(list(skydb.getDefValues('Dec:*' + source + '*').values()))
                    if len(ra) == 0 or len(dec) == 0:
                        ra = np.nan
                        dec = np.nan
                        logging.error('Cannot find the source '+source+'. I leave NaNs.')
                    else:
                        ra = ra.mean()
                        dec = dec.mean()
                        logging.info('Found average direction for '+source+' at ra:'+str(ra)+' - dec:'+str(dec))
                vals.append([ra, dec])
        sourceTable.append(list(zip(*(dirs,vals))))

    logging.info("Total file size: "+str(int(h5parm.H.get_filesize()/1024./1024.))+" M.")

    # Add CREATE entry to history and print summary of tables if verbose
    soltabs = solset.getSoltabs()
    for st in soltabs:
        if globaldbFile is None:
            st.addHistory('CREATE (by H5parm_importer.py from %s:%s/%s)' % (socket.gethostname(), os.path.abspath(''), "manual list"))
        else:
            st.addHistory('CREATE (by H5parm_importer.py from %s:%s/%s)' % (socket.gethostname(), os.path.abspath(''), globaldbFile))
    if verbose:
        logging.info(str(h5parm))

    del h5parm
    logging.info('Done.')    
    



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from losoto.lib_operations import *
from losoto._logging import logger as logging

logging.debug('Loading DELETEAXIS module.')

def _run_parser(soltab, parser, step):
    axisDelete = parser.getstr( step, 'axisDelete' ) # no default
    fromCell = parser.getstr( step, 'fromCell', '' )

    parser.checkSpelling( step, soltab, ['axisDelete', 'fromCell'])
    return run(soltab, axisDelete, fromCell)

def run( soltab, axisDelete, fromCell):
    """
    Delete an axis only keeping the values of a certain slice.

    Parameters
    ----------
    axisDelete : str
        Axis to delete.

    fromCell : str
        A cell value in axisDelete from which to keep the data values. If it is the string
        "first"/"last" then uses the first/last element of the axis.
    """
    import numpy as np

    if not axisDelete in soltab.getAxesNames():
        logging.error('Cannot find axis %s.' % axisDelete)
        return 1

    if fromCell == 'first':
        fromCell = soltab.getAxisValues(axisDelete)[0]
    elif fromCell == 'last':
        fromCell = soltab.getAxisValues(axisDelete)[-1]

    axisType = type(soltab.getAxisValues(axisDelete)[0])
    try:
        fromCell = np.array([fromCell]).astype(axisType)[0]
    except:
        logging.error('Cannot convert to type %s the value in fromCell: %s.' % (str(axisType),fromCell))
        return 1

    if not fromCell in soltab.getAxisValues(axisDelete):
        logging.error('Cannot find %s in %s.' % (fromCell, axisDelete))
        return 1

    logging.info("Delete axis on soltab: "+soltab.name)

    # get slice with 1 value to replicate
    soltab.setSelection(**{axisDelete:fromCell})
    axisDeleteIdx = soltab.getAxesNames().index(axisDelete)
    fromCellIdx = list(soltab.getAxisValues(axisDelete)).index(fromCell)
    vals = soltab.getValues(retAxesVals=False)
    vals = np.take(soltab.getValues(retAxesVals=False), fromCellIdx, axisDeleteIdx)
    weight = np.take(soltab.getValues(retAxesVals=False, weight=True), fromCellIdx, axisDeleteIdx)
    solset = soltab.getSolset()
    sttype = soltab.getType()
    stname = soltab.name

    axes = soltab.getAxesNames()
    axes.remove(axisDelete)
    axesVals = [soltab.getAxisValues(ax) for ax in axes]
    soltab.delete()

    st = solset.makeSoltab(sttype, stname, axesNames=axes, axesVals=axesVals, vals=vals, weights=weight)

    st.addHistory('DELETEAXIS (over axis %s from cell %s)' % (axisDelete, str(fromCell)))
    return 0



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Fill a solution table with beam response

from losoto.lib_operations import *
from losoto._logging import logger as logging

logging.debug('Loading LOFARBEAM module.')

# this funct is called by losoto to set parameters and call the real run()
def _run_parser(soltab, parser, step):
    ms = parser.getstr( step, 'ms') # no default
    inverse = parser.getbool( step, 'inverse', False)
    useElementResponse = parser.getbool( step, 'useElementResponse', True )
    useArrayFactor = parser.getbool( step, 'useArrayFactor', True )
    useChanFreq = parser.getbool( step, 'useChanFreq', True )

    parser.checkSpelling( step, soltab, ['ms', 'inverse', 'useElementResponse', 'useArrayFactor', 'useChanFreq'])
    return run(soltab, ms, inverse, useElementResponse, useArrayFactor, useChanFreq)

def run( soltab, ms, inverse=False, useElementResponse=True, useArrayFactor=True, useChanFreq=True ):
    """
    Generic unspecified step for easy expansion.

    Parameters
    ----------
    opt1 : float
        Is a mandatory parameter.

    """
    # load specific libs
    import numpy as np
    import casacore.tables as pt
    from lofar.stationresponse import stationresponse

    sr = stationresponse(ms, inverse, useElementResponse, useArrayFactor, useChanFreq)

    numants = pt.taql('select gcount(*) as numants from '+ms+'::ANTENNA').getcol('numants')[0]
    times = soltab.getAxisValues('time')

    for vals, coord, selection in soltab.getValuesIter(returnAxes=['ant','time','pol','freq'], weight=False):
        vals = reorderAxes( vals, soltab.getAxesNames(), ['ant','time','freq','pol'])

        for stationnum in range(numants):
            logging.debug('Working on station number %i' % stationnum)
            for itime, time in enumerate(times):
                beam = sr.evaluateStation(time=time, station=stationnum)
                # Reshape from [nfreq, 2, 2] to [nfreq, 4]
                beam = beam.reshape(beam.shape[0], 4)

                if soltab.getAxisLen('pol') == 2:
                    beam = beam[:,[0,3]] # get only XX and YY

                if soltab.getType() == 'amplitude':
                    vals[stationnum, itime, :, :] = np.abs(beam)
                elif soltab.getType() == 'phase':
                    vals[stationnum, itime, :, :] = np.angle(beam)
                else:
                    logging.error('Beam prediction work only for amplitude/phase solution tables.')
                    return 1

        vals = reorderAxes( vals, ['ant','time','freq','pol'], [ax for ax in soltab.getAxesNames() if ax in ['ant','time','freq','pol']] )
        soltab.setValues(vals, selection)

    return 0
 
 
